import meteva
import pandas as pd
import numpy as np
import datetime

def between_value_range(sta,start_value,end_value,start_open = False,end_open = False):
    sta1 = sta
    data_names = meteva.base.basicdata.get_stadata_names(sta)
    for data_name in data_names:
        if start_open:
            if end_open:
                sta1 = sta1.loc[(sta1[data_name] > start_value) & (sta1[data_name] < end_value)]
            else:
                sta1 = sta1.loc[(sta1[data_name] > start_value) & (sta1[data_name] <= end_value)]
        else:
            if end_open:
                sta1 = sta1.loc[(sta1[data_name] >= start_value) & (sta1[data_name] < end_value)]
            else:
                sta1 = sta1.loc[(sta1[data_name] >= start_value) & (sta1[data_name] <= end_value)]
    return sta1
#为站点数据中dataframe重新赋列名

def not_IV(sta):
    data_names = meteva.base.get_stadata_names(sta)
    sta1 = sta.loc[sta[data_names[0]] != meteva.base.IV]
    for i in range(1,len(data_names),1):
        sta1 = sta1.loc[sta1[data_names[i]] != meteva.base.IV]
    return sta1

def not_equal_to(sta,dele_value):
    data_names = meteva.base.get_stadata_names(sta)
    sta1 = sta.loc[sta[data_names[0]] != dele_value]
    for i in range(1,len(data_names),1):
        sta1 = sta1.loc[sta1[data_names[i]] != dele_value]
    return sta1

#为拥有多元素值的站点数据，在最后依次增加要素值的列表名
def in_member_list(data,member_list,name_or_index = "name"):
    if not isinstance(member_list,list) and not isinstance(member_list,np.ndarray):
        member_list = [member_list]
    if isinstance(data, pd.DataFrame):
        data_names = meteva.base.get_stadata_names(data)
        member_name_list = []
        if name_or_index == "name":
            for member in member_list:
                if member not in data_names:
                    print("error infomation: "+member +" is not a data column")
                    return
            member_name_list = member_list
        else:
            data_names = meteva.base.get_stadata_names(data)
            for member in member_list:
                member_name_list.append(data_names[member])
        columns = ['level', 'time', 'dtime', 'id', 'lon', 'lat'] + member_name_list
        sta1 = data.loc[:,columns]
        return sta1
    else:
        grid0 = meteva.base.basicdata.get_grid_of_data(data)
        num_list = []
        if name_or_index == "name":
            member_name_list = member_list
            for member_name in member_list:
                if member_name not in grid0.members:
                    print(member_name +" not exist in griddata's member_list")
                else:
                    for i in range(len(grid0.members)):
                        if member_name == grid0.members[i]:
                            num_list.append(i)
                            break

        else:
            member_name_list = []
            num_list = member_list
            for num in member_list:
                if num >= len(grid0.members):
                    print("网格数据member维度的size小于" + str(num))
                else:
                    member_name_list.append(grid0.members[num])
        dat = data.values[num_list, :, :, :, :, :]
        grid1 = meteva.base.basicdata.grid(grid0.glon, grid0.glat,
                                                            grid0.gtime, grid0.dtimes, grid0.levels,
                                                            member_list=member_name_list)
        grd1 = meteva.base.basicdata.grid_data(grid1, dat)
        return grd1

#为拥有多level层的站点数据，依次增加level层所表示的list列表

#为拥有多level层的站点数据，依次增加level层所表示的list列表
def in_level_list(data,level_list):
    if not isinstance(level_list,list) and not isinstance(level_list,np.ndarray):
        level_list = [level_list]

    if isinstance(data, pd.DataFrame):
        sta1 = data.loc[data['level'].isin(level_list)]
        return sta1
    else:
        grid0 = meteva.base.basicdata.get_grid_of_data(data)
        num_list = []
        level_list1 = []
        for level in level_list:
            if level not in grid0.levels:
                print(level +" not exist in griddata's level_list")
            else:
                for i in range(len(grid0.levels)):
                    if level == grid0.levels[i]:
                        num_list.append(i)
                        level_list1.append(level)
                        break

        dat = data.values[:, num_list, :, :, :, :]

        grid1 = meteva.base.basicdata.grid(grid0.glon, grid0.glat,
                                                            grid0.gtime, grid0.dtimes,level_list = level_list1,
                                                            member_list = grid0.members)
        grd1 = meteva.base.basicdata.grid_data(grid1, dat)
        return grd1

#为拥有多id的站点数据，依次增加id所表示的list列表
def in_id_list(sta,id_list):
    if not isinstance(id_list,list) and not isinstance(id_list,np.ndarray):
        id_list = [id_list]
    sta1 = sta.loc[sta['id'].isin(id_list)]
    return sta1


def in_lon_list(sta,lon_list):
    if not isinstance(lon_list,list) and not isinstance(lon_list,np.ndarray):
        lon_list = [lon_list]
    sta1 = sta.loc[sta['lon'].isin(lon_list)]
    return sta1

def in_lat_list(sta,lat_list):
    if not isinstance(lat_list,list) and not isinstance(lat_list,np.ndarray):
        lat_list = [lat_list]
    sta1 = sta.loc[sta['lat'].isin(lat_list)]
    return sta1

#为拥有多time层的站点数据，依次增加time层所表示的list列表

#为拥有多time层的站点数据，依次增加time层所表示的list列表
def in_time_list(data,time_list):

    if not isinstance(time_list,list) and not isinstance(time_list,np.ndarray):
        time_list = [time_list]
    time_list1 = []
    for time0 in time_list:
        time_list1.append(meteva.base.tool.time_tools.all_type_time_to_time64(time0))

    if isinstance(data, pd.DataFrame):
        sta1 = data.loc[data['time'].isin(time_list1)]
        return sta1
    else:
        grid0 = meteva.base.basicdata.get_grid_of_data(data)
        times_all = data["time"].values
        index_time_dict = {}
        for i in range(len(times_all)):
            time_str = meteva.base.tool.time_tools.all_type_time_to_str(times_all[i])
            index_time_dict[time_str] = i
        time_list1_valid = []
        index_time= []
        for j in range(len(time_list1)):
            time_str =  meteva.base.tool.time_tools.all_type_time_to_str(time_list1[j])
            if time_str in index_time_dict.keys():
                index_time.append(index_time_dict[time_str])
                time_list1_valid.append(time_list1[j])
        if len(index_time) == 0:return None
        dat = data.values[:, :,index_time,  :, :, :]
        times_fo = list(set(time_list1_valid))
        times_fo.sort()
        times_fo = np.array(times_fo)
        if times_fo.size == 1:
            gtime = [times_fo[0]]
        else:

            dhs_fo = times_fo[1] - times_fo[0]
            gtime = [times_fo[0], times_fo[-1], dhs_fo]

        grid1 = meteva.base.basicdata.grid(grid0.glon, grid0.glat,
                    gtime= gtime,
                    dtime_list=grid0.dtimes,level_list=grid0.levels,member_list=grid0.members)

        grd1 = meteva.base.basicdata.grid_data(grid1, dat)
        return grd1

def in_time_list_old(data,time_list):

    if not isinstance(time_list,list) and not isinstance(time_list,np.ndarray):
        time_list = [time_list]
    time_list1 = []
    for time0 in time_list:
        time_list1.append(meteva.base.tool.time_tools.all_type_time_to_time64(time0))

    if isinstance(data, pd.DataFrame):
        sta1 = data.loc[data['time'].isin(time_list1)]
        return sta1
    else:
        grid0 = meteva.base.basicdata.get_grid_of_data(data)
        times_all = data["time"].values
        index_time_dict = {}
        for i in range(len(times_all)):
            time_str = meteva.base.tool.time_tools.all_type_time_to_str(times_all[i])
            index_time_dict[time_str] = i
        index_time= []
        for j in range(len(time_list1)):
            time_str =  meteva.base.tool.time_tools.all_type_time_to_str(time_list1[j])
            if time_str in index_time_dict.keys():
                index_time.append(index_time_dict[time_str])
        if len(index_time) == 0:return None
        dat = data.values[:, :,index_time,  :, :, :]
        times_fo = list(set(time_list1))
        times_fo.sort()
        times_fo = np.array(times_fo)
        if times_fo.size == 1:
            gtime = [times_fo[0]]
        else:

            dhs_fo = times_fo[-1] - times_fo[0]
            gtime = [times_fo[0], times_fo[-1], dhs_fo]

        grid1 = meteva.base.basicdata.grid(grid0.glon, grid0.glat,
                    gtime= gtime,
                    dtime_list=grid0.dtimes,level_list=grid0.levels,member_list=grid0.members)

        grd1 = meteva.base.basicdata.grid_data(grid1, dat)
        return grd1


#为拥有多year的站点数据，依次增加year所表示的list列表
def in_year_list(sta,year_list):
    if not isinstance(year_list,list) and not isinstance(year_list,np.ndarray):
        year_list = [year_list]
    fo_times = pd.Series(0, index=sta['time'])
    sta1 = sta.loc[fo_times.index.year.isin(year_list)]
    return sta1

#为拥有多month的站点数据，依次增加month所表示的list列表
def in_month_list(sta,month_list):
    if not isinstance(month_list,list) and not isinstance(month_list,np.ndarray):
        month_list = [month_list]
    fo_times = pd.Series(0, index=sta['time'])
    sta1 = sta.loc[fo_times.index.month.isin(month_list)]
    return sta1

#为拥有多xun的站点数据，依次增加xun所表示的list列表
def in_xun_list(sta,xun_list):
    if not isinstance(xun_list,list) and not isinstance(xun_list,np.ndarray):
        xun_list = [xun_list]
    fo_times = pd.Series(0, index=sta['time'])
    mons = fo_times.index.month.astype(np.int16)
    days = fo_times.index.day.astype(np.int16)
    xuns = np.ceil(days / 10).values.astype(np.int16)
    xuns[xuns>3] = 3
    xuns += (mons - 1) * 3
    xuns = pd.Series(xuns)
    sta1 = sta.loc[xuns.isin(xun_list)]
    return sta1

#为拥有多hou的站点数据，依次增加hou所表示的list列表
def in_hou_list(sta,hou_list):
    if not isinstance(hou_list,list) and not isinstance(hou_list,np.ndarray):
        hou_list = [hou_list]
    fo_times = pd.Series(0, index=sta['time'])
    mons = fo_times.index.month.astype(np.int16)
    days = fo_times.index.day.astype(np.int16)
    hous = np.ceil(days / 5).values.astype(np.int16)
    hous[hous>6] = 6
    hous += (mons - 1) * 6
    hous = pd.Series(hous)
    sta1 = sta.loc[hous.isin(hou_list)]
    return sta1

#为拥有多day的站点数据，依次增加day所表示的list列表
def in_day_list(sta,day_list):
    if not isinstance(day_list,list) and not isinstance(day_list,np.ndarray):
        day_list = [day_list]
    days_list = []
    time0 = datetime.datetime(1900,1,1,0,0)
    seconds = 3600*24
    for day0 in day_list:
        if isinstance(day0,str):
            day0 = meteva.base.tool.time_tools.all_type_time_to_datetime(day0)
        day = (day0 - time0).total_seconds()//seconds
        days_list.append(day)
    indexs = (sta['time'] - time0)//np.timedelta64(1,"D")
    sta1 = sta.loc[indexs.isin(days_list)]
    return sta1

def in_dayofyear_list(sta,dayofyear_list):
    if not isinstance(dayofyear_list,list) and not isinstance(dayofyear_list,np.ndarray):
        dayofyear_list = [dayofyear_list]
    fo_times = pd.Series(0, index=sta['time'])
    sta1 = sta.loc[fo_times.index.dayofyear.isin(dayofyear_list)]
    return sta1

#为拥有多hour的站点数据，依次增加hour所表示的list列表
def in_hour_list(sta,hour_list):
    if not isinstance(hour_list,list) and not isinstance(hour_list,np.ndarray):
        hour_list = [hour_list]
    fo_times = pd.Series(0, index=sta['time'])
    sta1 = sta.loc[fo_times.index.hour.isin(hour_list)]
    return sta1

#按照时间中的（日期+小时）进行数据提取，不区分分钟
def in_day_hour_list(sta,day_hour_list):
    if not isinstance(day_hour_list, list) and not isinstance(day_hour_list, np.ndarray):
        day_hour_list = [day_hour_list]
    day_hours_list = []
    time0 = datetime.datetime(1900, 1, 1, 0, 0)
    seconds = 3600
    for day0 in day_hour_list:
        if isinstance(day0, str):
            day0 = meteva.base.tool.time_tools.all_type_time_to_datetime(day0)
        day = (day0 - time0).total_seconds() // seconds
        day_hours_list.append(day)
    indexs = (sta['time'] - time0) // np.timedelta64(1, "H")
    sta1 = sta.loc[indexs.isin(day_hours_list)]
    return sta1

#按照时间中的（年份+月份）进行数据提取
#为拥有多month的站点数据，依次增加month所表示的list列表
def in_year_month_list(sta,year_month_list):

    if not isinstance(year_month_list,list) and not isinstance(year_month_list,np.ndarray):
        ym_list = [year_month_list]
    fo_times = pd.Series(0, index=sta['time'])
    mons = fo_times.index.month.astype(np.int16)
    years = fo_times.index.year.astype(np.int16)
    yms = years * 100 + mons
    yms = pd.Series(yms)
    sta1 = sta.loc[yms.isin(ym_list)]
    return sta1

#为拥有多hour的站点数据，依次增加hour所表示的list列表
def in_minute_list(sta,minute_list):
    if not isinstance(minute_list,list) and not isinstance(minute_list,np.ndarray):
        minute_list = [minute_list]
    fo_times = pd.Series(0, index=sta['time'])
    sta1 = sta.loc[fo_times.index.minute.isin(minute_list)]
    return sta1


def between_time_range(data,start_time,end_time):
    if isinstance(data, pd.DataFrame):
        start_time = meteva.base.all_type_time_to_time64(start_time)
        end_time = meteva.base.all_type_time_to_time64(end_time)
        sta1 = data.loc[(data['time'] >= start_time) & (data['time'] <= end_time)]
        return sta1
    else:
        times = data["time"].values
        start_time1 = meteva.base.all_type_time_to_time64(start_time)
        end_time1 = meteva.base.all_type_time_to_time64(end_time)
        index = np.where((times >= start_time1) & (times <= end_time1))[0]
        grd1 = data.isel(time=index)
        return grd1

############
#为拥有多time层的站点数据，依次增加time层所表示的list列表
def in_ob_time_list(sta,time_list):
    if not isinstance(time_list,list) and not isinstance(time_list,np.ndarray):
        time_list = [time_list]
    time_list1 = []
    for time0 in time_list:
        time_list1.append(meteva.base.tool.time_tools.all_type_time_to_time64(time0))

    dtime_units ="hour"
    if len(sta.attrs) >0:
        if "dtime_units" in sta.attrs:
            if sta.attrs["dtime_units"] != "hour":
                dtime_units = "minute"
    if dtime_units == "hour":
        dtimes = sta["dtime"] * np.timedelta64(1, 'h')
    else:
        dtimes = sta["dtime"] * np.timedelta64(1, 'm')

    obtimes = sta['time'] + dtimes
    sta1 = sta.loc[obtimes.isin(time_list1)]
    return sta1

#为拥有多year的站点数据，依次增加year所表示的list列表
def in_ob_year_list(sta,year_list):
    if not isinstance(year_list,list) and not isinstance(year_list,np.ndarray):
        year_list = [year_list]
    dtime_units ="hour"
    if len(sta.attrs) >0:
        if "dtime_units" in sta.attrs:
            if sta.attrs["dtime_units"] != "hour":
                dtime_units = "minute"
    if dtime_units == "hour":
        dtimes = sta["dtime"] * np.timedelta64(1, 'h')
    else:
        dtimes = sta["dtime"] * np.timedelta64(1, 'm')
    obtimes = pd.Series(0,index = sta['time'] + dtimes)
    sta1 = sta.loc[obtimes.index.year.isin(year_list)]
    return sta1

#为拥有多month的站点数据，依次增加month所表示的list列表
def in_ob_month_list(sta,month_list):
    if not isinstance(month_list,list) and not isinstance(month_list,np.ndarray):
        month_list = [month_list]
    dtime_units ="hour"
    if len(sta.attrs) >0:
        if "dtime_units" in sta.attrs:
            if sta.attrs["dtime_units"] != "hour":
                dtime_units = "minute"
    if dtime_units == "hour":
        dtimes = sta["dtime"] * np.timedelta64(1, 'h')
    else:
        dtimes = sta["dtime"] * np.timedelta64(1, 'm')
    obtimes = pd.Series(0,index = sta['time'] + dtimes)
    sta1 = sta.loc[obtimes.index.month.isin(month_list)]
    return sta1


#为拥有多xun的站点数据，依次增加xun所表示的list列表
def in_ob_xun_list(sta,xun_list):
    if not isinstance(xun_list,list) and not isinstance(xun_list,np.ndarray):
        xun_list = [xun_list]
    dtime_units ="hour"
    if len(sta.attrs) >0:
        if "dtime_units" in sta.attrs:
            if sta.attrs["dtime_units"] != "hour":
                dtime_units = "minute"
    if dtime_units == "hour":
        dtimes = sta["dtime"] * np.timedelta64(1, 'h')
    else:
        dtimes = sta["dtime"] * np.timedelta64(1, 'm')
    obtimes = pd.Series(0,index = sta['time'] + dtimes)
    mons = obtimes.index.month.astype(np.int16)
    days = obtimes.index.day.astype(np.int16)
    xuns = np.ceil(days / 10).values.astype(np.int16)
    xuns[xuns>3] = 3
    xuns += (mons - 1) * 3
    xuns = pd.Series(xuns)
    sta1 = sta.loc[xuns.isin(xun_list)]
    return sta1

#为拥有多hou的站点数据，依次增加hou所表示的list列表
def in_ob_hou_list(sta,hou_list):
    if not isinstance(hou_list,list) and not isinstance(hou_list,np.ndarray):
        hou_list = [hou_list]
    dtime_units ="hour"
    if len(sta.attrs) >0:
        if "dtime_units" in sta.attrs:
            if sta.attrs["dtime_units"] != "hour":
                dtime_units = "minute"
    if dtime_units == "hour":
        dtimes = sta["dtime"] * np.timedelta64(1, 'h')
    else:
        dtimes = sta["dtime"] * np.timedelta64(1, 'm')
    obtimes = pd.Series(0,index = sta['time'] + dtimes)
    mons = obtimes.index.month.astype(np.int16)
    days = obtimes.index.day.astype(np.int16)
    hous = np.ceil(days / 5).values.astype(np.int16)
    hous[hous>6] = 6
    hous += (mons - 1) * 6
    hous = pd.Series(hous)
    sta1 = sta.loc[hous.isin(hou_list)]
    return sta1

#为拥有多day的站点数据，依次增加day所表示的list列表
def in_ob_dayofyear_list(sta,dayofyear_list):
    if not isinstance(dayofyear_list,list) and not isinstance(dayofyear_list,np.ndarray):
        dayofyear_list = [dayofyear_list]
    dtime_units ="hour"
    if len(sta.attrs) >0:
        if "dtime_units" in sta.attrs:
            if sta.attrs["dtime_units"] != "hour":
                dtime_units = "minute"
    if dtime_units == "hour":
        dtimes = sta["dtime"] * np.timedelta64(1, 'h')
    else:
        dtimes = sta["dtime"] * np.timedelta64(1, 'm')
    obtimes = pd.Series(0,index = sta['time'] + dtimes)
    sta1 = sta.loc[obtimes.index.dayofyear.isin(dayofyear_list)]
    return sta1

def in_ob_day_list(sta,day_list):
    if not isinstance(day_list,list) and not isinstance(day_list,np.ndarray):
        day_list = [day_list]
    dtime_units ="hour"
    if len(sta.attrs) >0:
        if "dtime_units" in sta.attrs:
            if sta.attrs["dtime_units"] != "hour":
                dtime_units = "minute"
    if dtime_units == "hour":
        dtimes = sta["dtime"] * np.timedelta64(1, 'h')
    else:
        dtimes = sta["dtime"] * np.timedelta64(1, 'm')
    obtimes = sta['time'] + dtimes
    days_list = []
    time0 = datetime.datetime(1900, 1, 1, 0, 0)
    seconds = 3600 * 24
    for day0 in day_list:
        if isinstance(day0,str):
            day0 = meteva.base.tool.time_tools.all_type_time_to_datetime(day0)
        day = (day0 - time0).total_seconds() // seconds
        days_list.append(day)
    indexs = (obtimes - time0) // np.timedelta64(1, "D")
    sta1 = sta.loc[indexs.isin(days_list)]
    return sta1



#为拥有多hour的站点数据，依次增加hour所表示的list列表
def in_ob_hour_list(sta,hour_list):
    if not isinstance(hour_list,list) and not isinstance(hour_list,np.ndarray):
        hour_list = [hour_list]
    dtime_units ="hour"
    if len(sta.attrs) >0:
        if "dtime_units" in sta.attrs:
            if sta.attrs["dtime_units"] != "hour":
                dtime_units = "minute"
    if dtime_units == "hour":
        dtimes = sta["dtime"] * np.timedelta64(1, 'h')
    else:
        dtimes = sta["dtime"] * np.timedelta64(1, 'm')
    obtimes = pd.Series(0,index = sta['time'] + dtimes)
    sta1 = sta.loc[obtimes.index.hour.isin(hour_list)]
    return sta1

#为拥有多minute的站点数据，依次增加hour所表示的list列表
def in_ob_minute_list(sta,minute_list):
    if not isinstance(minute_list,list) and not isinstance(minute_list,np.ndarray):
        minute_list = [minute_list]
    dtime_units ="hour"
    if len(sta.attrs) >0:
        if "dtime_units" in sta.attrs:
            if sta.attrs["dtime_units"] != "hour":
                dtime_units = "minute"
    if dtime_units == "hour":
        dtimes = sta["dtime"] * np.timedelta64(1, 'h')
    else:
        dtimes = sta["dtime"] * np.timedelta64(1, 'm')
    obtimes = pd.Series(0,index = sta['time'] + dtimes)
    sta1 = sta.loc[obtimes.index.minute.isin(minute_list)]
    return sta1

#按照时间中的（日期+小时）进行数据提取，不区分分钟
def in_ob_day_hour_list(sta,day_hour_list):
    if not isinstance(day_hour_list, list) and not isinstance(day_hour_list, np.ndarray):
        day_hour_list = [day_hour_list]
    day_hours_list = []
    time0 = datetime.datetime(1900, 1, 1, 0, 0)
    seconds = 3600
    for day0 in day_hour_list:
        if isinstance(day0, str):
            day0 = meteva.base.tool.time_tools.all_type_time_to_datetime(day0)
        day = (day0 - time0).total_seconds() // seconds
        day_hours_list.append(day)

    dtime_units ="hour"
    if len(sta.attrs) >0:
        if "dtime_units" in sta.attrs:
            if sta.attrs["dtime_units"] != "hour":
                dtime_units = "minute"
    if dtime_units == "hour":
        dtimes = sta["dtime"] * np.timedelta64(1, 'h')
    else:
        dtimes = sta["dtime"] * np.timedelta64(1, 'm')
    obtimes = pd.Series(0,index = sta['time'] + dtimes)

    indexs = (obtimes - time0) // np.timedelta64(1, "H")
    sta1 = sta.loc[indexs.isin(day_hours_list)]
    return sta1

#按照时间中的（年份+月份）进行数据提取
#为拥有多month的站点数据，依次增加month所表示的list列表
def in_ob_year_month_list(sta,year_month_list):

    if not isinstance(year_month_list,list) and not isinstance(year_month_list,np.ndarray):
        ym_list = [year_month_list]

    dtime_units ="hour"
    if len(sta.attrs) >0:
        if "dtime_units" in sta.attrs:
            if sta.attrs["dtime_units"] != "hour":
                dtime_units = "minute"
    if dtime_units == "hour":
        dtimes = sta["dtime"] * np.timedelta64(1, 'h')
    else:
        dtimes = sta["dtime"] * np.timedelta64(1, 'm')
    obtimes = pd.Series(0,index = sta['time'] + dtimes)

    #fo_times = pd.Series(0, index=sta['time'])
    mons = obtimes.index.month.astype(np.int16)
    years = obtimes.index.year.astype(np.int16)
    yms = years * 100 + mons
    yms = pd.Series(yms)
    sta1 = sta.loc[yms.isin(ym_list)]
    return sta1


def between_ob_time_range(sta,start_time,end_time):
    start_time = meteva.base.tool.time_tools.all_type_time_to_time64(start_time)
    end_time = meteva.base.tool.time_tools.all_type_time_to_time64(end_time)

    dtime_units ="hour"
    if len(sta.attrs) >0:
        if "dtime_units" in sta.attrs:
            if sta.attrs["dtime_units"] != "hour":
                dtime_units = "minute"
    if dtime_units == "hour":
        dtimes = sta["dtime"] * np.timedelta64(1, 'h')
    else:
        dtimes = sta["dtime"] * np.timedelta64(1, 'm')
    obtimes = sta['time'] + dtimes
    sta1 = sta.loc[(obtimes  >= start_time) & (obtimes  <= end_time)]
    return sta1
############

#为拥有多dtime的站点数据，依次增加dtime所表示的list列表
def in_dtime_list(data,dtime_list):
    if not isinstance(dtime_list,list) and not isinstance(dtime_list,np.ndarray):
        dtime_list = [dtime_list]
    if isinstance(data, pd.DataFrame):
        sta1 = data.loc[data['dtime'].isin(dtime_list)]
        return sta1
    else:
        grid0 = meteva.base.basicdata.get_grid_of_data(data)
        num_list = []
        dtime_list1 = []
        for dtime in dtime_list:
            if dtime not in grid0.dtimes:
                print(str(dtime) +" not exist in griddata's dtime_list")
            else:
                for i in range(len(grid0.dtimes)):
                    if dtime == grid0.dtimes[i]:
                        num_list.append(i)
                        dtime_list1.append(dtime)
                        break

        dat = data.values[:, :, :, num_list, :, :]

        grid1 = meteva.base.basicdata.grid(grid0.glon, grid0.glat,
                                                            grid0.gtime, dtime_list=dtime_list1,level_list = grid0.levels,
                                                            member_list = grid0.members)
        grd1 = meteva.base.basicdata.grid_data(grid1, dat)
        return grd1

#为拥有多dday的站点数据，依次增加dday所表示的list列表
def in_dday_list(sta,dday_list):
    if not isinstance(dday_list,list) and not isinstance(dday_list,np.ndarray):
        dday_list = [dday_list]
    days = np.ceil(sta['dtime'] / 24)
    sta1 = sta.loc[days.isin(dday_list)]
    return sta1

#为拥有多dhour的站点数据，依次增加dhour所表示的list列表
def in_dhour_list(sta,dhour_list):
    if not isinstance(dhour_list,list) and not isinstance(dhour_list,np.ndarray):
        dhour_list = [dhour_list]
    hours = sta['dtime']% 24
    sta1 = sta.loc[hours.isin(dhour_list)]
    return sta1

#为拥有多dminute的站点数据，依次增加minute所表示的list列表
def in_dminute_list(sta,dminute_list):
    if not isinstance(dminute_list,list) and not isinstance(dminute_list,np.ndarray):
        dminute_list = [dminute_list]
    minutes = sta['dtime'].map(lambda x: x - 10000)
    sta1 = sta.loc[minutes.isin(dminute_list)]
    return sta1

#返回的dtime在start_dtime和end_dtime之间
def between_dtime_range(data,start_dtime,end_dtime):
    if isinstance(data, pd.DataFrame):
        sta1 = data.loc[(data['dtime']>=start_dtime) & (data['dtime']<= end_dtime)]
        return sta1
    else:
        dtimes = data["dtime"].values
        index = np.where((dtimes>= start_dtime)&(dtimes<=end_dtime))[0]
        grd1 = data.isel(dtime=index)
        return grd1
#返回的lon在slon和elon之间
def between_lon_range(sta,slon,elon):
    sta1 = sta.loc[(sta['lon']>=slon) & (sta['lon']<= elon)]
    return sta1


#返回的lat在slat和elat之间
def between_lat_range(sta,slat,elat):
    sta1 = sta.loc[(sta['lat']>=slat) & (sta['lat']<= elat)]
    return sta1

#返回的alt在salt和ealt之间
def between_level_range(sta,slevel,elevel):
    sta1 = sta.loc[(sta['level']>=slevel) & (sta['level']<= elevel)]
    return sta1

#返回站点经纬度正好落在格点中的站点信息

def in_grid(sta,grid):
    sta1 = sta
    if grid.stime != grid.etime:
        sta1 = between_time_range(sta1,grid.stime,grid.etime)
    if len(grid.levels)>1:
        levels = np.array(grid.levels)
        levels.sort()
        sta1 = between_level_range(sta1, levels[0], levels[-1])
    if len(grid.dtimes)>1:
        dtimes = np.array(grid.dtimes)
        dtimes.sort()
        sta1 = between_dtime_range(sta1, dtimes[0], dtimes[-1])

    if grid.nlon>1:
        sta1 = between_lon_range(sta1, grid.slon, grid.elon)
    if grid.nlat>1:
        sta1 = between_lat_range(sta1, grid.slat, grid.elat)

    return sta1


def in_grid_xy(sta,grid):
    sta1 = between_lon_range(sta,grid.slon,grid.elon)
    sta2 = between_lat_range(sta1,grid.slat,grid.elat)
    return sta2

def in_grid_xyz(sta,grid):
    sta1 = in_grid_xy(sta,grid)
    levels = np.array(grid.levels)
    levels.sort()
    sta2 = between_level_range(sta1,levels[0],levels[-1])
    return sta2


def in_last_list(sta,last_list,drop_last = True):
    if not isinstance(last_list,list) and not isinstance(last_list,np.ndarray):
        last_list = [last_list]
    columns = list(sta.columns)
    sta1 = sta.loc[sta[columns[-1]].isin(last_list)]
    if drop_last:
        sta1 = sta1.loc[:, columns[0:-1]]
    return sta1

def between_last_range(sta,last_min,last_max,drop_last = True):
    columns = list(sta.columns)
    sta1 = sta.loc[(sta[columns[-1]] >= last_min) & (sta[columns[-1]] <= last_max)]
    if drop_last:
        sta1 = sta1.loc[:, columns[0:-1]]
    return  sta1


def between_one_column_value_range(sta,column_name,vmin,vmax):
    sta1 = sta.loc[(sta[column_name] >= vmin) & (sta[column_name] <= vmax)]
    return  sta1

def in_one_column_value_list(sta,column_name,value_list):
    sta1 = sta.loc[sta[column_name].isin(value_list)]
    return sta1


def by_stadata(sta,loc_sta):
    '''

    :param sta:
    :param loc_sta:
    :return:
    '''
    sta_combine = meteva.base.combine_expand_IV(sta,loc_sta)
    columns = list(sta_combine.columns)
    sta_sele = sta_combine.loc[:,columns[0:-1]]
    return sta_sele

def by_ob_stadata(sta,ob_stadata):
    '''

    :param sta:
    :param loc_sta:
    :return:
    '''
    data_names = meteva.base.get_stadata_names(sta)
    sta_combine = meteva.base.combine_on_obTime_id(ob_stadata,sta,need_match_ob=True)
    columns = list(sta_combine.columns)
    columns1 = columns[0:6]
    columns1.extend(columns[7:])
    sta_sele = sta_combine.loc[:,columns1]
    meteva.base.set_stadata_names(sta_sele,data_name_list=data_names)
    return sta_sele

def in_province_list(sta,province_name_list):
    if not isinstance(province_name_list,list):
        province_name_list = [province_name_list]
    ids = list(set(sta["id"].values))
    sta_province_name = meteva.base.tool.get_station_format_province_set(ids)
    sta_with_province_name = meteva.base.combine_expand_IV(sta, sta_province_name)
    sta1 = sta_with_province_name.loc[sta_with_province_name['province_name'].isin(province_name_list)]
    sta1 = sta1.drop(['province_name'], axis=1)

    return sta1

def not_nan(sta):
    danames = meteva.base.get_stadata_names(sta)
    # sta1 = sta.copy()
    # for name in danames:
    #     sta1 = sta[pd.notnull(sta1[name])]
    sta1 = sta.dropna(subset=danames)

    return sta1

#返回站点参数字典列表
def by_loc_dict(data,s):
    '''
    s 应具备如下样式
    s = {'member':["ecmwf","grapes"],列表形式
                'level' :[850,700],  列表形式,
                'time' : [datetime(2020,1,1,8,0)], 列表形式,
                "time_range":[datetime(2020,1,1,8,0),datetime(2020,1,2,8,0)], 闭区间,
                "year":[2020,2021],列表形式，
                "month": [1,2,3]，列表形式，
                "day": [datetime(2020,1,1,0,0)]列表形式，
                "dayofyear":[1,2,365,366] 列表形式，
                "hour":[0,1,23],列表形式，

                'ob_time' : [datetime(2020,1,1,8,0)], 列表形式,
                "ob_time_range":[datetime(2020,1,1,8,0),datetime(2020,1,2,8,0)], 闭区间,
                "ob_year":[2020,2021],列表形式，
                "ob_month": [1,2,3]，列表形式，
                "ob_day": [datetime(2020,1,1,0,0)]列表形式，
                "ob_dayofyear":[1,2,365,366] 列表形式，
                "ob_hour":[0,1,23],列表形式，

                'dtime':[24,36]，列表形式,
                'ddya': [0,1,2]，列表形式，
                "dhour":[0,1,23],列表形式，

                'id':[54511]， 列表形式,
                'lon': [70,140], 闭区间,
                'lat' :[10,60], 闭区间,
                "stadata": 站点形式数据
    }
    '''
    if isinstance(data,pd.DataFrame):
        if s is None:return data
        sta1 = data
        if "member" in s.keys():
            sta1 = in_member_list(sta1,s["member"])
        if "level" in s.keys():
            sta1 = in_level_list(sta1,s['level'])
        if "time" in s.keys():
            sta1 = in_time_list(sta1, s['time'])
        if "time_range" in s.keys():
            if not isinstance(s['time_range'],list):
                print("time_range参数需为列表形式的包含起始时刻和结束时刻的时间参数，时间可以是datetime格式，datetime64或者字符串形式（例如2019010108）")
            sta1 = between_time_range(sta1, s['time_range'][0],s['time_range'][1])
        if "year" in s.keys():
            sta1 = in_year_list(sta1,s["year"])
        if "month" in s.keys():
            sta1 = in_month_list(sta1,s["month"])
        if "day" in s.keys():
            sta1 = in_day_list(sta1,s["day"])
        if "date" in s.keys():
            sta1 = in_day_list(sta1,s["date"])
        if "dayofyear" in s.keys():
            sta1 = in_dayofyear_list(sta1,s["dayofyear"])
        if "hour" in s.keys():
            sta1 = in_hour_list(sta1, s["hour"])

        if "ob_time" in s.keys():
            sta1 = in_ob_time_list(sta1, s['ob_time'])
        if "ob_time_range" in s.keys():
            if not isinstance(s['ob_time_range'], list):
                print("ob_time_range参数需为列表形式的包含起始时刻和结束时刻的时间参数，时间可以是datetime格式，datetime64或者字符串形式（例如2019010108）")
            sta1 = between_ob_time_range(sta1, s['ob_time_range'][0],s['ob_time_range'][1])
        if "ob_year" in s.keys():
            sta1 = in_ob_year_list(sta1,s["ob_year"])
        if "ob_month" in s.keys():
            sta1 = in_ob_month_list(sta1,s["ob_month"])
        if "ob_day" in s.keys():
            sta1 = in_ob_day_list(sta1,s["ob_day"])
        if "ob_dayofyear" in s.keys():
            sta1 = in_ob_dayofyear_list(sta1,s["ob_dayofyear"])
        if "ob_hour" in s.keys():
            sta1 = in_ob_hour_list(sta1, s["ob_hour"])

        if "dtime" in s.keys():
            sta1 = in_dtime_list(sta1,s["dtime"])
        if "dtime_range" in s.keys():
            sta1 = between_dtime_range(sta1, s["dtime_range"][0],s["dtime_range"][1])
        if "dday" in s.keys():
            sta1 = in_dday_list(sta1,s["dday"])
        if "dhour" in s.keys():
            sta1 = in_dday_list(sta1,s["dhour"])
        if "lon" in s.keys():
            sta1 = between_lon_range(sta1,s["lon"][0],s["lon"][1])
        if "lat" in s.keys():
            sta1 = between_lat_range(sta1,s["lat"][0],s["lat"][1])
        if "id" in s.keys():
            sta1 = in_id_list(sta1,s["id"])


        return sta1



#返回站点参数字典列表
def sele_by_dict(data,s):
    '''
    :param data: [站点数据](https://www.showdoc.cc/nmc?page_id=3744334022014027)
    :param s:用于选择的数据的参数字典，它具备如下形式
    s = {    "member":成员的名称，同时提取多个时采用列表形式
                    "level":层次的名称，同时提取多个时采用列表形式
                    "time":时间（以起报时刻为准），可以是datetime,datetime64或“2019010108”类似的字符串形式，同时提取多个时采用列表形式
                    "time_range":时间范围，列表形式，第一个元素为起始时间，第二个为结束时间，时间可以是datetime,datetime64或“2019010108”类似的字符串形式
                    "year":要提取的数据的年份（以起报时刻为准），同时提取多个时采用列表形式
                    "month":要提取的数据的月份（以起报时刻为准），同时提取多个时采用列表形式
                    "day":要提取的数据的日期（以起报时刻为准），可以是datetime,datetime64或“20190101”类似的字符串形式，同时提取多个时采用列表形式
                    "dayofyear":要提取的数据在一年中的排序（以起报时刻为准），整数形式，同时提取多个时采用列表形式
                    "hour":要提取的数据的小时数（以起报时刻为准），0-23的整数，同时提取多个时采用列表形式
                    "ob_time":时间（以观测时刻为准），可以是datetime,datetime64或“2019010108”类似的字符串形式，同时提取多个时采用列表形式
                    "ob_time_range":时间范围，列表形式，第一个元素为起始时间，第二个为结束时间，时间可以是datetime,datetime64或“2019010108”类似的字符串形式
                    "ob_year":要提取的数据的年份（以观测时刻为准），同时提取多个时采用列表形式
                    "ob_month":要提取的数据的月份（以观测时刻为准），同时提取多个时采用列表形式
                    "ob_day":要提取的数据的日期（以观测时刻为准），可以是datetime,datetime64或“20190101”类似的字符串形式，同时提取多个时采用列表形式
                    "ob_dayofyear":要提取的数据在一年中的排序（以观测时刻为准），整数形式，同时提取多个时采用列表形式
                    "ob_hour":要提取的数据的小时数（以观测时刻为准），0-23的整数，同时提取多个时采用列表形式
                    "dtime":要提取的数据的时效，整数形式，同时提取多个是采用列表形式
                    "dtime_range":时间范围，列表形式，第一个元素为起始时效，第二个为结束时效
                    "dday":  要提取的数据的时效dtime整除以24的值，整数形式，同时提取多个时采用列表形式
                    "dhour":要提取的数据的时效dtime除24的余数，整数形式，同时提取多个时采用列表形式
                    "lon":要提取的数据的经度范围，列表形式，第一个元素为起始经度，第二个为结束经度
                    "lat":要提取的数据的纬度范围，列表形式，第一个元素为起始经度，第二个为结束经度
                    "id": 要提取的数据的站号，同时提取多个是采用列表形式
                    "gxy": 提取某个平面网格范围内的数据
                    "gxyz": 提取某个三维网格范围内的数据
                    "stadata": 对于stadata中level，time，dtime，id四个坐标中非缺省的部分，从data中提取和stadata坐标一致的站点数据
                    "ob_stadata": 将ob_stadata 和sta_all用combine_on_obtime_id匹配，保留的样本再删除ob_stadata
                    "value": 提取所有数据列都在给定取值范围的数据，列表形式第一个元素为数据最低值，第二个为数据最高值
                    "drop_IV": 该参数为True时，删除包含缺省值的行
                    "last_range":包含起始值和结束值的列表，取出最后一列取值在该取值范围的数据，并删除最后一列的数据
                    "last": 取出最后一列包含lastL的行，如何选择多个类型，lastL采用列表形式，并删除最后一列的数据

    }
    '''
    if s is None:return data




    p_set = ["member","level","time","time_range","year","month","day","dayofyear","hour", "ob_time","ob_time_range" ,"ob_year",
              "ob_month", "ob_day","ob_dayofyear","ob_hour","dtime","dtime_range","dday","dhour" ,
              "lon","lat", "id","grid","gxy", "gxyz" ,"stadata","ob_stadata","value","drop_IV","last" , "last_range","drop_last","province_name",
             "day_hour","year_month","ob_day_hour","ob_year_month"]


    sta1 = data

    if isinstance(data, pd.DataFrame):
        data_names = meteva.base.get_stadata_names(data)
        p_set.extend(data_names)
        data_names_range = []
        for data_name in data_names:
            data_names_range.append(str(data_name)+"_range")
        p_set.extend(data_names_range)
        for key in s.keys():
            if key in data_names and key not in ["lon","lat"]:                 #lon,lat 默认表示取值范围，而不是具体的值
                value_one = s[key]
                if not isinstance(value_one,list):
                    value_one = [value_one]
                sta1 = in_one_column_value_list(sta1,key, value_one)
            elif key in data_names_range:
                value_one = s[key]
                data_name = key[0:-6]
                sta1 = between_one_column_value_range(sta1,data_name,value_one[0],value_one[1])


    key_set = s.keys() #set(list(s.keys()))
    if(not set(p_set) >= key_set):
        print("参数s的字典中包含本程序不能识别的关键词")
        return None

    member = None
    if "member" in s.keys():
        member = s["member"]
        
    level = None
    if "level" in s.keys():
        level = s['level']
    
    time = None
    if "time" in s.keys():
        time = s['time']

    time_range = None
    if "time_range" in s.keys():
        time_range = s["time_range"]

    year = None
    if "year" in s.keys():
        year = s["year"]

    month = None
    if "month" in s.keys():
        month = s["month"]

    day = None
    if "day" in s.keys():
        day = s["day"]

    dayofyear = None
    if "dayofyear" in s.keys():
        dayofyear = s["dayofyear"]

    hour = None
    if "hour" in s.keys():
        hour = s["hour"]


    minute = None
    if "minute" in s.keys():
        minute = s["minute"]


    ob_time = None
    if "ob_time" in s.keys():
        ob_time = s['ob_time']

    ob_time_range = None
    if "ob_time_range" in s.keys():
        ob_time_range = s['ob_time_range']

    ob_year = None
    if "ob_year" in s.keys():
        ob_year = s["ob_year"]

    ob_month = None
    if "ob_month" in s.keys():
        ob_month = s["ob_month"]

    ob_day = None
    if "ob_day" in s.keys():
        ob_day = s["ob_day"]

    ob_dayofyear = None
    if "ob_dayofyear" in s.keys():
        ob_dayofyear = s["ob_dayofyear"]

    ob_hour = None
    if "ob_hour" in s.keys():
        ob_hour = s["ob_hour"]
    ob_minute = None
    if "ob_minute" in s.keys():
        ob_minute = s["ob_minute"]

    dtime = None
    if "dtime" in s.keys():
        dtime = s["dtime"]

    dtime_range = None
    if "dtime_range" in s.keys():
        dtime_range = s["dtime_range"]

    dday = None
    if "dday" in s.keys():
        dday = s["dday"]

    dhour = None
    if "dhour" in s.keys():
        dhour = s["dhour"]

    lon = None
    if "lon" in s.keys():
        lon = s["lon"]

    lat = None
    if "lat" in s.keys():
        lat = s["lat"]

    id = None
    if "id" in s.keys():
        id = s["id"]

    stadata = None
    if "stadata" in s.keys():
        stadata = s["stadata"]

    ob_stadata = None
    if "ob_stadata" in s.keys():
        ob_stadata = s["ob_stadata"]

    grid = None
    if "grid" in s.keys():
        grid = s["grid"]

    gxy = None
    if "gxy" in s.keys():
        gxy = s["gxy"]

    gxyz = None
    if "gxyz" in s.keys():
        gxyz = s["gxyz"]

    value = None
    if "value" in s.keys():
        value = s["value"]

    drop_IV = False
    if "drop_IV" in s.keys():
        drop_IV = s["drop_IV"]

    last = None
    if "last" in s.keys():
        last = s["last"]

    last_range = None
    if "last_range" in s.keys():
        last_range = s["last_range"]

    province_name = None
    if "province_name" in s.keys():
        province_name = s["province_name"]

    day_hour = None
    if "day_hour" in s.keys():
        day_hour = s["day_hour"]

    year_month = None
    if "year_month" in s.keys():
        year_month = s["year_month"]

    ob_day_hour = None
    if "ob_day_hour" in s.keys():
        ob_day_hour = s["ob_day_hour"]

    ob_year_month = None
    if "ob_year_month" in s.keys():
        ob_year_month = s["ob_year_month"]

    drop_last = True
    if "drop_last" in s.keys():
        drop_last = s["drop_last"]


    sta1 = sele_by_para(sta1,member,level,time,time_range,year,month,day,dayofyear,hour,minute,ob_time,ob_time_range,ob_year,ob_month,ob_day,ob_dayofyear,
                 ob_hour,ob_minute,dtime,dtime_range,dday,dhour,lon,lat,id,grid,gxy,gxyz,stadata,value,drop_IV,last,last_range,province_name,drop_last,ob_stadata,
                        day_hour,year_month,ob_day_hour,ob_year_month)
    return sta1


def sele_by_para(data,member = None,level = None,time = None,time_range = None,year = None,month = None,day = None,dayofyear = None,hour = None,minute = None,
           ob_time=None, ob_time_range=None, ob_year=None, ob_month=None, ob_day=None, ob_dayofyear=None, ob_hour=None,ob_minute = None,
           dtime = None,dtime_range = None,dday = None, dhour = None,lon = None,lat = None,id = None,grid = None,gxy = None,gxyz = None,stadata = None,
                 value = None,drop_IV = False,last = None,last_range = None,province_name = None,drop_last = True,ob_stadata = None,day_hour=None,year_month=None,
                 ob_day_hour=None,ob_year_month=None,**kwargs):
    '''
    :param data: [站点数据](https://www.showdoc.cc/nmc?page_id=3744334022014027)
    :param member:成员的名称，同时提取多个时采用列表形式
    :param level:层次的名称，同时提取多个时采用列表形式
    :param time:时间（以起报时刻为准），可以是datetime,datetime64或“2019010108”类似的字符串形式，同时提取多个时采用列表形式
    :param time_range:时间范围，列表形式，第一个元素为起始时间，第二个为结束时间，时间可以是datetime,datetime64或“2019010108”类似的字符串形式
    :param year:要提取的数据的年份（以起报时刻为准），同时提取多个时采用列表形式
    :param month:要提取的数据的月份（以起报时刻为准），同时提取多个时采用列表形式
    :param day:要提取的数据的日期（以起报时刻为准），可以是datetime,datetime64或“20190101”类似的字符串形式，同时提取多个时采用列表形式
    :param dayofyear:要提取的数据在一年中的排序（以起报时刻为准），整数形式，同时提取多个时采用列表形式
    :param hour:要提取的数据的小时数（以起报时刻为准），0-23的整数，同时提取多个时采用列表形式
    :param ob_time:时间（以观测时刻为准），可以是datetime,datetime64或“2019010108”类似的字符串形式，同时提取多个时采用列表形式
    :param ob_time_range:时间范围，列表形式，第一个元素为起始时间，第二个为结束时间，时间可以是datetime,datetime64或“2019010108”类似的字符串形式
    :param ob_year:要提取的数据的年份（以观测时刻为准），同时提取多个时采用列表形式
    :param ob_month:要提取的数据的月份（以观测时刻为准），同时提取多个时采用列表形式
    :param ob_day:要提取的数据的日期（以观测时刻为准），可以是datetime,datetime64或“20190101”类似的字符串形式，同时提取多个时采用列表形式
    :param ob_dayofyear:要提取的数据在一年中的排序（以观测时刻为准），整数形式，同时提取多个时采用列表形式
    :param ob_hour:要提取的数据的小时数（以观测时刻为准），0-23的整数，同时提取多个时采用列表形式
    :param dtime:要提取的数据的时效，整数形式，同时提取多个是采用列表形式
    :param dtime_range:时间范围，列表形式，第一个元素为起始时效，第二个为结束时效
    :param dday:  要提取的数据的时效dtime整除以24的值，整数形式，同时提取多个时采用列表形式
    :param dhour: 要提取的数据的时效dtime除24的余数，整数形式，同时提取多个时采用列表形式
    :param lon:要提取的数据的经度范围，列表形式，第一个元素为起始经度，第二个为结束经度
    :param lat:要提取的数据的纬度范围，列表形式，第一个元素为起始经度，第二个为结束经度
    :param id: 要提取的数据的站号，同时提取多个是采用列表形式
    :param grid: 网格信息类变量，提取多维矩形网格范围内的数据，grid中size>1的维度的坐标范围会被用作选择的已经，size=1的维度会被忽略
    :param gxy: 网格信息类变量，采用其中经纬度范围提取水平矩形网格范围内的数据
    :param gxyz: 网格信息类变量，采用其中经纬度和层次范围提取三维矩形网格范围内的数据
    :param stadata: 站点数据类变量，对于stadata中level，time，dtime，id四个坐标中非缺省的部分，从data中提取和stadata坐标一致的站点数据
    :param value: 提取所有数据列都在给定取值范围的数据，列表形式，第一个元素为数据最低值，第二个为数据最高值
    :param drop_IV: 该参数为True时，删除包含缺省值的行
    :param last_range: 包含起始值和结束值的列表，取出最后一列取值在该取值范围的数据，并删除最后一列的数据
    :param last: 取出最后一列包含lastL的行，如何选择多个类型，lastL采用列表形式，并删除最后一列的数据
    :return:  [站点数据](https://www.showdoc.cc/nmc?page_id=3744334022014027)
    '''
    sta1 = data
    if isinstance(data, pd.DataFrame):
        data_names = meteva.base.get_stadata_names(data)
        data_names_range = []
        for data_name in data_names:
            data_names_range.append(str(data_name)+"_range")
        for key in kwargs.keys():
            if key in data_names:
                value_one = kwargs[key]
                if not isinstance(value_one,list):
                    value_one = [value_one]
                sta1 = in_one_column_value_list(sta1,key, value_one)
            elif key in data_names_range:
                value_one = kwargs[key]
                data_name = key[0:-6]
                sta1 = between_one_column_value_range(sta1,data_name,value_one[0],value_one[1])
            else:
                print("输入的参数"+key+"不是数据选取函数可接受的参数")

    if member is not None:
        sta1 = in_member_list(sta1,member)
    if level is not None:
        sta1 = in_level_list(sta1,level)
    if time is not None:
        sta1 = in_time_list(sta1, time)
    if time_range is not None:
        if not isinstance(time_range,list):
            print("time_range参数需为列表形式的包含起始时刻和结束时刻的时间参数，时间可以是datetime格式，datetime64或者字符串形式（例如2019010108）")
        sta1 = between_time_range(sta1, time_range[0],time_range[1])
    if year is not None:
        sta1 = in_year_list(sta1,year)
    if month is not None:
        sta1 = in_month_list(sta1,month)
    if day is not None:
        sta1 = in_day_list(sta1,day)
    if dayofyear is not None:
        sta1 = in_dayofyear_list(sta1,dayofyear)
    if hour is not None:
        sta1 = in_hour_list(sta1, hour)
    if minute is not None:
        sta1 = in_minute_list(sta1, minute)

    if ob_time is not None:
        sta1 = in_ob_time_list(sta1, ob_time)
    if ob_time_range is not None:
        if not isinstance(ob_time_range,list):
            print("ob_time_range参数需为列表形式的包含起始时刻和结束时刻的时间参数，时间可以是datetime格式，datetime64或者字符串形式（例如2019010108）")
        sta1 = between_ob_time_range(sta1, ob_time_range[0],ob_time_range[1])
    if ob_year is not None:
        sta1 = in_ob_year_list(sta1,ob_year)
    if ob_month is not None:
        sta1 = in_ob_month_list(sta1,ob_month)
    if ob_day is not None:
        sta1 = in_ob_day_list(sta1,ob_day)
    if ob_dayofyear is not None:
        sta1 = in_ob_dayofyear_list(sta1,ob_dayofyear)
    if ob_hour is not None:
        sta1 = in_ob_hour_list(sta1, ob_hour)
    if ob_minute is not None:
        sta1 = in_ob_minute_list(sta1, ob_minute)
    if dtime is not None:
        sta1 = in_dtime_list(sta1,dtime)
    if dtime_range is not None:
        if not isinstance(dtime_range,list):
            print("dtime_range参数需为列表形式的包含起始时效（整数）和结束时效（整数）的参数")
        sta1 = between_dtime_range(sta1, dtime_range[0],dtime_range[1])
    if dday is not None:
        sta1 = in_dday_list(sta1,dday)
    if dhour is not None:
        sta1 = in_dhour_list(sta1,dhour)
    if lon is not None:
        if not isinstance(lon,list):
            lon = [lon - 1e-6, lon + 1e-6]
        elif len(lon)==1:
            lon = [lon[0]-1e-6,lon[0]+1e-6]

            #print("lon参数需为列表形式的包含起始经度（浮点数）和结束经度（浮点）的参数")
        sta1 = between_lon_range(sta1,lon[0],lon[1])
    if lat is not None:
        if not isinstance(lat,list):
            lat = [lat - 1e-6, lat + 1e-6]
        elif len(lat) == 1:
            lat = [lat[0] - 1e-6, lat[0] + 1e-6]
            #print("lat参数需为列表形式的包含起始纬度（浮点数）和结束纬度（浮点）的参数")
        sta1 = between_lat_range(sta1,lat[0],lat[1])


    if id is not None:
        sta1 = in_id_list(sta1,id)
    if grid is not None:
        sta1 = in_grid(sta1,grid)
    if gxy is not None:
        sta1 = in_grid_xy(sta1,gxy)
    if gxyz is not None:
        sta1 = in_grid_xyz(sta1,gxyz)
    if stadata is not None:
        sta1 = by_stadata(sta1,stadata)
    if value is not None:
        if not isinstance(value, list):
            print("value参数需为列表形式的包含起始值和结束值的参数")
        sta1 = between_value_range(sta1,value[0],value[1])

    if ob_stadata is not None:
        sta1 = by_ob_stadata(sta1,ob_stadata)

    if drop_IV is True:
        sta1 = not_IV(sta1)
        sta1 = not_nan(sta1)
    if last_range is not None:
        sta1 = between_last_range(sta1,last_range[0],last_range[1],drop_last)
    if last is not None:
        sta1 = in_last_list(sta1,last,drop_last)
    if province_name is not None:
        sta1 = in_province_list(sta1,province_name)

    if day_hour is not None:
        sta1 = in_day_hour_list(sta1,day_hour)
    if year_month is not None:
        sta1 = in_year_month_list(sta1,year_month)
    if ob_day_hour is not None:
        sta1 = in_ob_day_hour_list(sta1,ob_day_hour)
    if ob_year_month is not None:
        sta1 = in_ob_year_month_list(sta1,ob_year_month)

    return sta1



def drop_by_para(data,member = None,level = None,time = None,time_range = None,year = None,month = None,day = None,dayofyear = None,hour = None,minute = None,
           ob_time=None, ob_time_range=None, ob_year=None, ob_month=None, ob_day=None, ob_dayofyear=None, ob_hour=None,ob_minute = None,
           dtime = None,dtime_range = None,dday = None, dhour = None,lon = None,lat = None,id = None,grid = None,gxy = None,gxyz = None,stadata = None,
                 value = None,drop_IV = False,last = None,last_range = None,province_name = None,drop_last = True,ob_stadata = None,
                 day_hour = None,year_month = None,ob_day_hour = None,ob_year_month = None,**kwargs):

    data.reset_index(drop=True, inplace=True)
    data_drop =  sele_by_para(data,member,level,time,time_range,year,month,day,dayofyear,hour,minute,ob_time,ob_time_range,ob_year,ob_month,ob_day,ob_dayofyear,
                 ob_hour,ob_minute,dtime,dtime_range,dday,dhour,lon,lat,id,grid,gxy,gxyz,stadata,value,drop_IV,last,last_range,province_name,drop_last,ob_stadata
                ,day_hour,year_month,ob_day_hour,ob_year_month)
    data_left = data.drop(data_drop.index)

    return data_left


from meteva.base.fun import *
from meteva.product.program.fun import *
def plot(sta_ob_and_fos0,method,s = None,g = None,gll = None,save_dir = None,save_path = None,title =None,**kwargs):
    if s is not None:
        if g is not None:
            if g == "last_range" or g == "last_step":
                s["drop_last"] = False
            else:
                s["drop_last"] = True

    sta_ob_and_fos = meteva.base.sele_by_para(sta_ob_and_fos0, drop_IV=True)
    sta_ob_and_fos = sele_by_dict(sta_ob_and_fos, s)
    #discription_uni = get_unique_coods(sta_ob_and_fos)
    sta_ob_and_fos_list,gll1 = group(sta_ob_and_fos,g,gll)
    data_name = meteva.base.get_stadata_names(sta_ob_and_fos_list[0])

    group_num = len(sta_ob_and_fos_list)
    #title = None
    #if "title" in kwargs.keys():
    #    title = kwargs["title"]

    valid_gll = []
    if type(method) == str:
        method =  globals().get(method)


    if isinstance(title, list):
        if group_num != len(title):
            print("手动设置的title数目和要绘制的图形数目不一致")
            return

    if save_path is not None:
        if isinstance(save_path,str):
            save_path = [save_path]
        if group_num != len(save_path):
            print("手动设置的save_path数目和要绘制的图形数目不一致")
            return

    for i in range(group_num):
        sta = sta_ob_and_fos_list[i]
        if(len(sta.index) == 0):
            pass
        else:
            valid_group_list = None
            if gll1 is None:
                valid_gll = None
            else:
                valid_group_list = gll1[i]
                valid_gll.append(gll1[i])

            #ob = sta[data_name[0]].values
            #fo = sta[data_name[1:]].values.T


            if isinstance(title, list):
                title1 = title[i]
            else:
                if title is None:
                    title1 = meteva.product.program.get_title_from_dict(method, s, g, valid_group_list, None)
                else:
                    title1 = meteva.product.program.get_title_from_dict(title, s, g, valid_group_list, None)
            if save_path is None:
                #title2 = title1.replace("\n", "").replace(":", "")
                save_path1 = get_save_path(save_dir,method,g,valid_group_list,"",".png","")
            else:
                save_path1 = save_path[i]

            kwargs["title"] = title1
            method_para = method.__code__.co_varnames[:method.__code__.co_argcount]

            meteva.base.creat_path(save_path1)
            if method.__name__.find("_uv")>=0:
                if "member_list" in method_para:
                    member_list = data_name[2::2]
                    member_list1 = []
                    for member in member_list:
                        str1 = member.replace("u_","")
                        member_list1.append(str1)
                    kwargs["member_list"] = member_list1
                u_ob = sta[data_name[0]].values
                v_ob = sta[data_name[1]].values
                u_fo = sta[data_name[2::2]].values.T
                v_fo = sta[data_name[3::2]].values.T
                method(u_ob,u_fo,v_ob,v_fo,save_path = save_path1,**kwargs)
            else:
                if "member_list" in method_para and "member_list" not in kwargs.keys():
                    kwargs["member_list"] = data_name[1:]
                ob = sta[data_name[0]].values
                fo = sta[data_name[1:]].values.T
                method(ob, fo,save_path = save_path1,**kwargs)

    return valid_gll


def plot1(sta_ob_and_fos0,method,s = None,g = None,gll = None,save_dir = None,title = None,para1 = None):

    if s is not None:
        if g is not None:
            if g == "last_range" or g == "last_step":
                s["drop_last"] = False
            else:
                s["drop_last"] = True
    sta_ob_and_fos = sele_by_dict(sta_ob_and_fos0, s)
    #discription_uni = get_unique_coods(sta_ob_and_fos)
    sta_ob_and_fos_list,gll1 = group(sta_ob_and_fos,g,gll)
    data_name = meteva.base.get_stadata_names(sta_ob_and_fos)
    fo_num = len(data_name) -1
    ensemble_score_method = [meteva.method.box_plot_ensemble,meteva.method.rank_histogram]
    group_num = len(sta_ob_and_fos_list)

    valid_gll = []
    if type(method) == str:
        method =  globals().get(method)
    if method in ensemble_score_method:
        for i in range(group_num):
            sta = sta_ob_and_fos_list[i]
            if(len(sta.index) == 0):
                pass
            else:
                valid_group_list = None
                if gll1 is None:
                    valid_gll = None
                else:
                    valid_group_list = gll1[i]
                    valid_gll.append(gll1[i])
                ob = sta[data_name[0]].values
                fo = sta[data_name[1:]].values
                save_path = get_save_path(save_dir,method,g,valid_group_list,"",".png",title)
                #title1 = get_title(method,g,valid_group_list,"",title,discription_uni)

                title1 = meteva.product.program.get_title_from_dict(method, s, g, valid_group_list,
                                                                    None)
                if para1 is None:
                    method(ob, fo,save_path = save_path,title = title1)
                else:
                    method(ob, fo,para1,save_path = save_path,title = title1)
    else:
        for i in range(group_num):
            sta = sta_ob_and_fos_list[i]
            if(len(sta.index) == 0):
                pass
            else:
                valid_group_list = None
                if gll1 is None:
                    valid_gll = None
                else:
                    valid_group_list = gll1[i]
                    valid_gll.append(gll1[i])
                ob = sta[data_name[0]].values

                for j in range(fo_num):
                    fo = sta[data_name[j+1]].values
                    save_path = get_save_path(save_dir,method,g,valid_group_list,data_name[j+1],".png",title)
                    #title1 = get_title(method,g,valid_group_list,data_name[j+1],title,discription_uni)
                    title1 = meteva.product.program.get_title_from_dict(method, s, g, valid_group_list,data_name[j+1])

                    if para1 is None:
                        method(ob, fo,save_path = save_path,title = title1)
                    else:
                        method(ob, fo,para1,save_path = save_path,title = title1)


    return valid_gll



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Feb 28 15:50:52 2023

@author: tangbuxing
"""
import sys
import numpy as np
import matplotlib.pyplot as plt
import meteva
from  matplotlib import  cm


def Geo_Box_Plot(frequency_ob,frequency_fo,grade_list,member_list = None):
    out = {}
    if len(frequency_ob) != len(grade_list) :
        sys.exit("参数frequency_ob与参数grade_list长度需一样。")
        return
    if len(frequency_fo) != len(grade_list) :
        sys.exit("参数frequency_fo与参数grade_list长度需一样。")
        return



    if member_list is None:
        xticks = ['观测']
        if member_list[0] == 1:
            xticks.append('预报')
        else:
            for i in range(member_list[0]):
                xticks.append('预报' + str(i + 1))
    else:
        xticks = member_list


    obs_array = np.array([])
    for i in range(len(grade_list)):
        c = np.repeat(grade_list[i],frequency_ob[i])
        #print(c)
        obs_array = np.append(obs_array, c)
        #print(d)

    fst_array = np.array([])
    for i in range(len(grade_list)):
        c = np.repeat(grade_list[i],frequency_fo[i])
        #print(c)
        fst_array= np.append(fst_array, c)
        #print(d)
    out = plt.boxplot(np.array([obs_array,fst_array]), labels=xticks)



    plt.show()
    plt.close()
    return


def GeoBoxPlot(x, areas):
    out = {}
    if len(x) != len(areas):
        sys.exit("参数x与参数areas长度需一样。")
    d = np.array([])

    for i in range(len(x)):
        c = np.repeat(x[i],areas[i])
        #print(c)
        d = np.append(d, c)
        #print(d)
    out = plt.boxplot(d)
    plt.show()
    plt.close()
    return
    
if __name__ == '__main__':

    #x = [4,9,1,3,10,6,7]
    #areas = [1,1,1,1,2,1,3]
    #GeoBoxPlot(x, areas)
    import meteva.base as meb
    path_ob = r'H:\test_data\input\mem\mode\ob\rain03\20072611.000.nc'
    path_fo = r'H:\test_data\input\mem\mode\ec\rain03\20072608.003.nc'
    grid1 = meb.grid([100, 120, 0.05], [24, 40, 0.05])
    grd_ob = meb.read_griddata_from_nc(path_ob, grid=grid1, time="2020072611", dtime=0, data_name="OBS")
    grd_fo = meb.read_griddata_from_nc(path_fo, grid=grid1, time="2020072608", dtime=3, data_name="ECMWF")
    obs_array = grd_ob.values.squeeze()
    fst_array = grd_fo.values.squeeze()

    conf_mx = meteva.method.multi_category.table.frequency_table(obs_array,fst_array,grade_list=np.arange(0,300))

    print(conf_mx)











import xarray as xr
import os
from multiprocessing import Process
import sys
import copy


class Logger(object):
    def __init__(self, filename='default.log', stream=sys.stdout):
        self.terminal = stream
        self.log = open(filename, 'w')

    def write(self, message):
        self.terminal.write(message)
        self.log.write(message)

    def flush(self):
        pass



def open_dataset(filename,filter_by_keys={}):
    sys.stdout = Logger('grib_read_out.txt', sys.stdout)
    sys.stderr = Logger('grib_read_err.txt', sys.stderr)
    backend_kwargs = {"filter_by_keys":filter_by_keys, "indexpath": ""}
    ds1 = xr.open_dataset(filename, engine="cfgrib", backend_kwargs=backend_kwargs)
    print(ds1)
    ds1.close()




def pro(filename,filter_by_keys):
    tmpp = Process(target=open_dataset, kwargs={"filename":filename,"filter_by_keys":filter_by_keys})
    tmpp.start()
    tmpp.join()

def print_grib_file_info(filename,filter_by_keys = {},info = None,print_info = True):
    pro(filename,filter_by_keys)
    if info is None:
        info = []

    has_error=False
    if os.path.exists("grib_read_err.txt"):
        grib_read_err = open("grib_read_err.txt", "r")
        err_infos = grib_read_err.read()
        strs = err_infos.split("\n")
        if len(strs)>1:
            has_error=True
    if os.path.exists("grib_read_out.txt"):
        grib_read_out = open("grib_read_out.txt", "r")
        out_infos = grib_read_out.read()

        if len(out_infos)>0:
            print("*************************************************************************")
            print("使用参数 filter_by_keys = "+str(filter_by_keys)+"查看到的数据内容为：")
            print(out_infos)

        if "typeOfLevel" not in filter_by_keys:
            valid_levelType = ["isobaricInhPa", "surface", "meanSea", "heightAboveGround", "heightAboveGroundLayer",
                               "cloudBase", "cloudTop", "pressureFromGroundLayer"]
            out_info = out_infos.split("\n")
            if len(out_info)>2:
                i_c = 0
                i_d = 0
                for i  in range(len(out_info)):
                    if out_info[i].strip() == "Coordinates:":
                        i_c = i
                    if out_info[i].strip() == "Data variables:":
                        i_d = i

                index1 = out_infos.find("Data variables:")
                index2 = out_infos.find("Attributes:")
                variables = out_infos[index1:index2]

                had_valid = False
                for i in range(i_c+1,i_d):
                    str1 = out_info[i].replace("*","")
                    str_list = str1.split()
                    str2 = str_list[0].strip()
                    #if str2 in valid_levelType:
                    if variables.find(str2)>=0 and str2 in valid_levelType:
                        had_valid = True
                        print("*************************************************************************")
                        print("请在读取该grib文件时添加参数 \nfilter_by_keys= {'typeOfLevel':'"+str2+"'}")

                if not had_valid:
                    print("请根据以上数据内容信息，确认所需读取的物理量对应的levelType")
        if not has_error:
            return

    if has_error:
        grib_read_err = open("grib_read_err.txt", "r")
        err_infos = grib_read_err.read()
        if err_infos.find("filter_by_keys")>=0:
            strs = err_infos.split("\n")
            for str1 in strs:
                if str1.find("filter_by_keys=") >= 0:
                    filter_by_keys1 = {}
                    str1s = str1.split("=")
                    str1s = str1s[1].replace("{","").replace("}","").replace("'","").replace(" ","").split(",")
                    for str2 in str1s:
                        key,value = str2.split(":")
                        filter_by_keys1[key] = value
                        info.append(filter_by_keys1)
                        try:
                            print_grib_file_info(filename,filter_by_keys=filter_by_keys1,info = info,print_info = False)
                        except:
                            pass
        elif err_infos.find("skipping variable: paramId=") >= 0:
            strs = err_infos.split("\n")
            for str1 in strs:
                if str1.find("skipping variable: paramId=") >= 0:
                    filter_by_keys1 = copy.deepcopy(filter_by_keys)
                    str1s = str1.split()
                    str1s_kv = str1s[2].split("==")
                    filter_by_keys1[str1s_kv[0]] = int(str1s_kv[1])
                    #str1s_kv = str1s[3].split("=")
                    #filter_by_keys1[str1s_kv[0]] = str1s_kv[1]
                    info.append(filter_by_keys1)
                    try:
                        print_grib_file_info(filename, filter_by_keys=filter_by_keys1, info=info, print_info=False)
                    except:
                        pass

    if print_info:
        info_short = []
        print("在读取数据时需添加如下参数当中的一个")
        for info1 in info:
            if "shortName" in info1.keys():
                info_short.append(info1)
            else:
                print("filter_by_keys = "+str(info1))
        for info1 in info_short:
            shortName = info1["shortName"]
            del info1["shortName"]
            print("若需查看变量"+shortName+"请添加参数:")
            print("filter_by_keys = "+str(info1))
    else:
        return info

if __name__=="__main__":

    #path = r"\\10.28.16.239\ensemble_datasets\ReanalysisData\CRA\2014\20140101\CRA40LAND_SURFACE_2014010112_GLB_0P25_HOUR_V1_0_0.grib"
    #print_grib_file_info(path)


    #print_grib_file_info(r"H:\test_data\input\meb\era5-levels-members.grib")
    print_grib_file_info(r"D:\book\test_data\charpter5\test.grib")



import time
import numpy as np
from meteva.method.space.fuzzy_logic.lib.datagrabber_spatialVx import datagrabber_spatialVx
from meteva.method.space.fuzzy_logic.lib.fss2dfun import fss2dfun
from meteva.method.space.fuzzy_logic.lib.fuzzyjoint2dfun import fuzzyjoint2dfun
from meteva.method.space.fuzzy_logic.lib.mincvg2dfun import mincvg2dfun
from meteva.method.space.fuzzy_logic.lib.multicon2dfun import multicon2dfun
from meteva.method.space.fuzzy_logic.lib.progmatic2dfun import pragmatic2dfun
from meteva.method.space.fuzzy_logic.lib.thresholder import thresholder
from meteva.method.space.fuzzy_logic.lib.thresholder_spatialVx import thresholder_spatialVx
from meteva.method.space.fuzzy_logic.lib.hoods2dPrep import hoods2dPrep
from meteva.method.space.fuzzy_logic.lib.hoods2dSetUpLists import hoods2dSetUpLists
from meteva.method.space.fuzzy_logic.lib.hoods2dsmooth import hoods2dsmooth
from meteva.method.space.fuzzy_logic.lib.make_spatialVx import make_spatialVx




def hoods2d(input_object, which_methods=None, time_point=1,
            obs=0, model=0, pe=None, levels=None, max_n=None, rule=">=", verbose=False):

    #hoods2d可以同时计算"mincvr", "multi_event", "fuzzy", "joint", "fss", "pragmatic" 六种空间方法的结果
    if which_methods is None:
        which_methods = ["mincvr", "multi_event", "fuzzy", "joint", "fss", "pragmatic"]


    smooth_fun = "hoods2dsmooth"  # 平滑方法
    smooth_params = None
    input_object['data'] = [input_object['X'], input_object['Xhat']]  #从输入的数据中提取观测和预报数据

    # 根据输入参数计算对应的levels, pe, max_n
    input_object = hoods2dPrep(input_object=input_object, pe=pe, levels=levels,
                               max_n=max_n, smooth_fun=smooth_fun, smooth_params=smooth_params)

    #是否显示
    if verbose:
        begin_time = time.time()

    thresholds = input_object['thresholds']  #输入的数据中包含阈值
    q = thresholds[0].shape[0]
    levels = input_object['levels']   #从输入中提取窗口宽度参数
    l = len(levels)

    # 抓取数据
    dat = datagrabber_spatialVx(input_object, time_point=time_point, obs=obs, model=model)

    x = dat['X'] #观测场
    outmat = np.zeros((l, q))
    #pd.DataFrame(outmat, columns=json.loads(input_object["qs"]), index=levels)

    # 根据输入参数，初始化返回结果
    out = hoods2dSetUpLists(input_object=input_object, which_methods=which_methods, mat=outmat)
    sub = input_object["subset"]


    if verbose:
        print("Looping through thresholds.\n")
    for threshold in range(0, q, 1):
        if len({"mincvr", "mincvr", "multi_event", "fuzzy", "joint", "fss", "pragmatic"}.intersection(
                which_methods)) > 0:
            if verbose:
                print("\n", "Setting up binary objects for threshold ", threshold, "\n")

            # 根据thresholds, rule，替换矩阵中的元素
            dat2 = thresholder_spatialVx(input_object, func_type="binary", th=threshold, rule=rule, time_point=time_point, obs=obs,
                                         model=model)
            ix = dat2['X']  #观测
            iy = dat2['Xhat'] #预报


        if "fss" in which_methods:
            if sub is None:
                f0 = np.mean(ix)
            else:
                f0 = np.mean(ix['subset'])
            if threshold == 1:
                out["fss"]["fss_uniform"] = 0.5 + f0 / 2
                out["fss"]["fss_random"] = f0
            else:
                out["fss"]["fss_uniform"] = [out["fss"]["fss_uniform"], 0.5 + f0 / 2]
                out["fss"]["fss_random"] = [out["fss"]["fss_random"], f0]
        if verbose:
            print("Looping through levels.\n")
        for level in range(0, l, 1):
            if verbose:
                print("Neighborhood length = ", levels[level], "\n")

            # 平滑处理
            level_w = hoods2dsmooth(x, levels[level], None, True)
            if len({"mincvr", "multi_event", "fuzzy", "joint", "pragmatic", "fss"}.intersection(
                    which_methods)) > 0:
                if len({"mincvr", "multi_event", "fuzzy", "joint", "fss"}.intersection(which_methods)) > 0:
                    s_px = hoods2dsmooth(ix, levels[level], level_w)
                s_py = hoods2dsmooth(iy, levels[level], level_w)

            if len({"mincvr", "multi_event"}.intersection(which_methods)) > 0:
                s_ix = thresholder(s_px, func_type="binary", th=input_object["pe"][level])
                s_iy = thresholder(s_py, func_type="binary", th=input_object["pe"][level])
                if "mincvr" in which_methods:
                    tmp = mincvg2dfun(s_iy=s_iy, s_ix=s_ix, subset=sub)
                    out["mincvr"]["pod"][level, threshold] = tmp["pod"]
                    out["mincvr"]["far"][level, threshold] = tmp["far"]
                    out["mincvr"]["ets"][level, threshold] = tmp["ets"]

                if "multi_event" in which_methods:
                    tmp = multicon2dfun(s_iy=s_iy, ix=ix, subset=sub)
                    out["multi_event"]["pod"][level, threshold] = tmp["pod"]
                    out["multi_event"]["f"][level, threshold] = tmp["f"]
                    out["multi_event"]["hk"][level, threshold] = tmp["hk"]

            if len({"fuzzy", "joint"}.intersection(which_methods)) > 0:
                tmp = fuzzyjoint2dfun(s_py=s_py, s_px=s_px, subset=sub)
                if "fuzzy" in which_methods:
                    out["fuzzy"]["pod"][level, threshold] = tmp["fuzzy"]["pod"]
                    out["fuzzy"]["far"][level, threshold] = tmp["fuzzy"]["far"]
                    out["fuzzy"]["ets"][level, threshold] = tmp["fuzzy"]["ets"]

                if "joint" in which_methods:
                    out["joint"]["pod"][level, threshold] = tmp["joint"]["pod"]
                    out["joint"]["far"][level, threshold] = tmp["joint"]["far"]
                    out["joint"]["ets"][level, threshold] = tmp["joint"]["ets"]

            if "fss" in which_methods:
                out["fss"]["values"][level, threshold] = fss2dfun(s_py=s_py, s_px=s_px, subset=sub)
            if "pragmatic" in which_methods:
                tmp = pragmatic2dfun(s_py=s_py, ix=ix)
                out["pragmatic"]["bs"][level, threshold] = tmp["bs"]
                out["pragmatic"]["bss"][level, threshold] = tmp["bss"]

    if verbose:
        print(time.time() - begin_time)
    #out["time_point"] = time_point
    #out["model_num"] = model

    return out


if __name__ == '__main__':
    import meteva.base as meb
    grid1 = meb.grid([100, 120, 0.05], [24, 40, 0.05])
    path_ob = r'H:\test_data\input\mem\mode\ob\rain03\20072611.000.nc'
    path_fo_03 = r'H:\test_data\input\mem\mode\ec\rain03\20072608.003.nc'
    path_fo_27 = r'H:\test_data\input\mem\mode\ec\rain03\20072508.027.nc'
    grd_ob = meb.read_griddata_from_nc(path_ob, grid=grid1, time="2020072611", dtime=0, data_name="OBS")
    grd_fo_03 = meb.read_griddata_from_nc(path_fo_03, grid=grid1, time="2020072608", dtime=3, data_name="ECMWF")
    grd_fo_27 = meb.read_griddata_from_nc(path_fo_27, grid=grid1, time="2020072508", dtime=27, data_name="ECMWF")

    obs_array = grd_ob.values.squeeze()
    fst_array = grd_fo_03.values.squeeze()
    hold = make_spatialVx(obs_array,fst_array)
    look = hoods2d(hold,levels=[3])
    print(look["joint"])


import numpy as np
import math
import meteva
import os
import traceback
import warnings

warnings.filterwarnings("ignore")



def write_griddata_to_micaps4(da,save_path = "a.txt",creat_dir = False,effectiveNum = 3,show = False,title = None,inte=None,vmin=None,vmax=None):

    """
    输出micaps4格式文件
    :param da:xarray多维数据信息,需要用 meteva 的格式
    :param save_path:存储路径
    :param creat_dir:存储路径中的文件夹若不存在是否创建
    :param effectiveNum：有效数字，默认 3
    :param show:是否输出存储结果，默认否
    :param title:MICAPS4第四类格式的title，默认根据 save_path 自动生成
    :param inte:MICAPS4第四类格式的等值线间隔，默认根据数值自动生成
    :param vmin:MICAPS4第四类格式的等值线起始值，默认根据数值自动生成
    :param vmax:MICAPS4第四类格式的等值线终止值，默认根据数值自动生成
    :return 最终按照需要保存的路径，将da数据保存为m4格式
    """
    try:
        dir = os.path.split(os.path.abspath(save_path))[0]
        if not os.path.isdir(dir):
            if not creat_dir:
                print("文件夹：" + dir + "不存在")
                return False
            else:
                meteva.base.tool.path_tools.creat_path(save_path)

        grid = meteva.base.basicdata.get_grid_of_data(da)
        nlon = grid.nlon
        nlat = grid.nlat
        slon = grid.slon
        slat = grid.slat
        elon = grid.elon
        elat = grid.elat
        dlon = grid.dlon
        dlat = grid.dlat
        level = grid.levels[0]
        stime = grid.stime_str
        year = stime[0:4]
        month = stime[4:6]
        day = stime[6:8]
        hour = stime[8:10]
        hour_range = str(grid.dtimes[0])
        values = da.values
        grid_values = np.squeeze(values)

        if(vmax is None):
            vmax = math.ceil(max(grid_values.flatten()))
        if(vmin is None):
            vmin = math.ceil(min(grid_values.flatten()))
        
        if(inte is None):
            dif = (vmax - vmin) / 10.0
            if dif ==0:
                inte = 1
            else:
                inte = math.pow(10, math.floor(math.log10(dif)))
            # 用基本间隔，将最大最小值除于间隔后小数点部分去除，最后把间隔也整数化
            r = dif / inte
            if r < 3 and r >= 1.5:
                inte = inte * 2
            elif r < 4.5 and r >= 3:
                inte = inte * 4
            elif r < 5.5 and r >= 4.5:
                inte = inte * 5
            elif r < 7 and r >= 5.5:
                inte = inte * 6
            elif r >= 7:
                inte = inte * 8
            vmin = inte * ((int)(vmin / inte) - 1)
            vmax = inte * ((int)(vmax / inte) + 1)

        end = len(save_path)
        start = max(0, end - 16)

        if title is None:
            title = ("diamond 4 " + save_path[start:end] + "\n"
                     + year + " " + month + " " + day + " " + hour + " " + hour_range + " " + str(level) + "\n"
                     + "{:.6f}".format(grid.dlon) + " " + "{:.6f}".format(grid.dlat) + " " + "{:.6f}".format(
                        grid.slon) + " " + "{:.6f}".format(grid.elon) + " "
                     + "{:.6f}".format(grid.slat) + " " + "{:.6f}".format(grid.elat) + " " + str(grid.nlon) + " " + str(
                        grid.nlat) + " "
                     + str(inte) + " " + str(vmin) + " " + str(vmax) + " 1 0")
        else:

            title = ("diamond 4 " + title + "\n"
                     + year + " " + month + " " + day + " " + hour + " " + hour_range + " " + str(level) + "\n"
                     + "{:.6f}".format(grid.dlon) + " " + "{:.6f}".format(grid.dlat) + " " + "{:.6f}".format(
                        grid.slon) + " " + "{:.6f}".format(grid.elon) + " "
                     + "{:.6f}".format(grid.slat) + " " + "{:.6f}".format(grid.elat) + " " + str(grid.nlon) + " " + str(
                        grid.nlat) + " "
                     + str(inte) + " " + str(vmin) + " " + str(vmax) + " 1 0")

        # 二维数组写入micaps文件
        format_str = "%." + str(effectiveNum) + "f "

        np.savetxt(save_path, grid_values, delimiter=' ',

                   fmt=format_str, header=title, comments='',encoding='GBK')
        if show:
            print('成功输出至' + save_path)
        return True

    except:
        exstr = traceback.format_exc()
        print(exstr)
        return False

    
    
def write_griddata_to_nc(da,save_path = "a.txt",creat_dir = False,effectiveNum = 3,show = False):
    try:
        dir = os.path.split(os.path.abspath(save_path))[0]
        if not os.path.isdir(dir):
            if not creat_dir:
                print("文件夹：" + dir + "不存在")
                return False
            else:
                meteva.base.tool.path_tools.creat_path(save_path)
        scale_factor = math.pow(10, -effectiveNum)
        # print(scale_factor)
        encodingdict = {da.name: {
            'dtype': 'int32',
            'scale_factor': scale_factor,
            'zlib': True,
            '_FillValue': None
        }
        }

        da.to_netcdf(save_path, encoding=encodingdict)
        if show:
            print('成功输出至' + save_path)
        return True
    except:
        exstr = traceback.format_exc()
        print(exstr)
        return False


def write_griddata_to_micaps11(wind, save_path="a.txt", creat_dir=False, effectiveNum=3, show=False, title=None):
    try:
        dir = os.path.split(os.path.abspath(save_path))[0]
        if not os.path.isdir(dir):
            if not creat_dir:
                print("文件夹：" + dir + "不存在")
                return False
            else:
                meteva.base.tool.path_tools.creat_path(save_path)
        grid0 = meteva.base.basicdata.get_grid_of_data(wind)
        nlon = grid0.nlon
        nlat = grid0.nlat
        slon = grid0.slon
        slat = grid0.slat
        elon = grid0.elon
        elat = grid0.elat
        dlon = grid0.dlon
        dlat = grid0.dlat
        level = grid0.levels[0]
        stime = grid0.stime_str
        year = stime[0:4]
        month = stime[4:6]
        day = stime[6:8]
        hour = stime[8:10]
        values = wind.values
        grid_values = np.squeeze(values).reshape(2 * nlat, nlon)

        end = len(save_path)
        start = max(0, end - 16)

        if title is None:
            title = ("diamond 11 " + save_path[start:end] + "\n"
                     + year + " " + month + " " + day + " " + hour + " " + str(level) + "\n"
                     + str(grid0.dlon) + " " + str(grid0.dlat) + " " + str(grid0.slon) + " " + str(grid0.elon) + " "
                     + str(grid0.slat) + " " + str(grid0.elat) + " " + str(grid0.nlon) + " " + str(grid0.nlat))
        else:

            title = ("diamond 11 " + title + "\n"
                     + year + " " + month + " " + day + " " + hour + " " + str(level) + "\n"
                     + str(grid0.dlon) + " " + str(grid0.dlat) + " " + str(grid0.slon) + " " + str(grid0.elon) + " "
                     + str(grid0.slat) + " " + str(grid0.elat) + " " + str(grid0.nlon) + " " + str(grid0.nlat))

        format_str = "%." + str(effectiveNum) + "f "

        np.savetxt(save_path, grid_values, delimiter=' ',
                   fmt=format_str, header=title, comments='')
        if show:
            print('成功输出至' + save_path)
        return True
    except:
        exstr = traceback.format_exc()
        print(exstr)
        return False


def tran_griddata_to_gds_flow(da):
    grid0 = meteva.base.get_grid_of_data(da)
    discriminator = b"mdfs"
    if len(grid0.members) == 1:
        data_type = 4
    elif len(grid0.members) == 2:
        data_type = 11
    else:
        print("仅支持micap4类和micaps11类数据输出成GDS格式")
    data_type_byte = np.ndarray.tobytes(np.array([data_type]).astype(np.int16))
    mName = grid0.members[0]
    mName = mName.encode(encoding='utf-8')
    if len(mName) < 20:
        mName = mName + np.ndarray.tobytes(np.zeros(20 - len(mName)).astype(np.int8))

    eleName = b""
    if "eleName" in da.attrs.keys():
        eleName = da.attrs["eleName"]
    if len(eleName) < 50:
        eleName = eleName + np.ndarray.tobytes(np.zeros(50 - len(eleName)).astype(np.int8))

    description = b""
    if "description" in da.attrs.keys():
        description = da.attrs["description"]
    if len(description) < 30:
        description = description + np.ndarray.tobytes(np.zeros(30 - len(description)).astype(np.int8))

    level = np.ndarray.tobytes(np.array(grid0.levels[0]).astype(np.float32))
    y_m_d_h_timezone_peroid = np.ndarray.tobytes(np.array([2021, 1, 1, 8, 8, 0]).astype(np.int32))
    slon_elon_dlon = np.ndarray.tobytes(np.array([grid0.slon, grid0.elon, grid0.dlon]).astype(np.float32))
    nlon = np.ndarray.tobytes(np.array([grid0.nlon]).astype(np.int32))
    slat_elat_dlat = np.ndarray.tobytes(np.array([grid0.slat, grid0.elat, grid0.dlat]).astype(np.float32))
    nlat = np.ndarray.tobytes(np.array([grid0.nlat]).astype(np.int32))
    vmin, vmax, inte = meteva.base.tool.plot_tools.get_isoline_set(da)
    sValue_eValue_dValue = np.ndarray.tobytes(np.array([vmin, vmax, inte]).astype(np.float32))
    blank = np.ndarray.tobytes(np.zeros(100).astype(np.int8))
    value_bytes = np.ndarray.tobytes(da.values.astype(np.float32))
    bytes1 = discriminator + data_type_byte + mName + eleName + description + level + y_m_d_h_timezone_peroid
    bytes2 = slon_elon_dlon + nlon + slat_elat_dlat + nlat + sValue_eValue_dValue + blank + value_bytes
    bytes = bytes1 + bytes2

    return bytes


def write_griddata_to_gds_file(da, save_path="a.txt", creat_dir=False, show=False):
    try:
        dir = os.path.split(os.path.abspath(save_path))[0]
        if not os.path.isdir(dir):
            if not creat_dir:
                print("文件夹：" + dir + "不存在")
                return False
            else:
                meteva.base.tool.path_tools.creat_path(save_path)

        bytes = tran_griddata_to_gds_flow(da)
        br = open(save_path, 'wb')
        br.write(bytes)
        br.close()
        if show:
            print('成功输出至' + save_path)
        return True
    except:
        exstr = traceback.format_exc()
        print(exstr)
        return False


if __name__ == "__main__":
    grd = meteva.base.read_griddata_from_micaps4(r"H:\test_data\input\meb\m4.txt")
    write_griddata_to_gds_file(grd,save_path=r"H:\test_data\output\meb\gds_test.000")




from .im import im
from .solutionset import solutionset
from .locperf import locperf


def locperfer(x1, x2, thresh, k):
    x1 = im(x1)
    x1["v"] = (x1["v"] >= thresh)
    ix1 = solutionset(x1)
    return locperf(X=x2, Y=ix1, which_stats = "ph", k = k)["ph"]



import numpy as np
#import cv2
import math
#import pyreadr
from scipy.ndimage import convolve
import meteva
import copy

def count_index(grd_ob,grd_fo,thresholds):
    '''

    :param grd_ob: 观测网格数据
    :param grd_fo: 预报网格数据
    :param threholds: 等级阈值
    :return:
    '''
    x = copy.deepcopy(grd_ob.values.squeeze())
    aindex_x = cindex(x,thresh=thresholds)
    y = copy.deepcopy(grd_fo.values.squeeze())
    aindex_y = cindex(y, thresh=thresholds)
    resulut = {"count_index":[aindex_x["count_index"],aindex_y["count_index"]],
               "count_nonzero":[aindex_x["count_nonzero"],aindex_y["count_nonzero"]],
               "count_connect":[aindex_x["count_connect"],aindex_y["count_connect"]]}
    return resulut


def cindex(x, thresh=None):
    if thresh is None:
        thresh = 1e-08

    sx = np.zeros(x.shape)
    sx[x >= thresh] = 1
    kernel =    np.array([[1, 1, 1],
                         [1, 1, 1],
                        [1, 1, 1]])/9
    sm =  convolve(sx, kernel) #均值滤波
    #sm = sx
    thresh1 = 50/255   #平滑后亮度大于50
    sm1 = sm >= thresh1
    NP = np.count_nonzero(sx)   #所有的非0格点
    label_array = meteva.method.mode.measure_label(sm1)  #转换成label矩阵
    num_obj =int(np.max(label_array))  #连通域的数量
    NC = num_obj - 1
    c =  1 - (NC - 1) / (math.sqrt(NP) + NC)
    result = {"count_index":c,"count_nonzero":NP,"count_connect":num_obj}

    return result

# def cindex_with_cv(x, thresh=None):
#     if thresh is None:
#         thresh = 1e-08
#     sx = x
#     sx[sx >= thresh] = 255
#     sx[sx < thresh] = 0
#     cv2.imwrite("pic/cindex/data.png", sx)
#     img = cv2.imread("pic/cindex/data.png")
#     gray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)  # 将图像转换成灰度
#     blur = cv2.blur(gray, (3, 3))  # 均值滤波
#     ret, thresh = cv2.threshold(blur, 50, 255, cv2.THRESH_BINARY)
#     NP = np.count_nonzero(sx)
#     num_obj, labels = cv2.connectedComponents(thresh)
#     NC = num_obj - 1
#     return 1 - (NC - 1) / (math.sqrt(NP) + NC)


if __name__ == '__main__':
    import meteva.base as meb
    import meteva.method as mem
    grid1 = meb.grid([100, 120, 0.05], [24, 40, 0.05])
    path_ob = r'H:\test_data\input\mem\mode\ob\rain03\20072611.000.nc'
    path_fo = r'H:\test_data\input\mem\mode\ec\rain03\20072608.003.nc'
    grd_ob = meb.read_griddata_from_nc(path_ob, grid=grid1, time="2020072611", dtime=0, data_name="OBS")
    grd_fo = meb.read_griddata_from_nc(path_fo, grid=grid1, time="2020072608", dtime=3, data_name="ECMWF")
    result = mem.space.count_index(grd_ob,grd_fo,thresholds=5)

    print(result)


import numpy as npy
import pandas as pd
import time
import math
import copy
from .utils import get_attributes_for_feat, remove_key_from_list
from .feature_comps import feature_comps
from .feature_axis import feature_axis


def interester(look, properties=None,
               weights=None, b1=None,
               b2=None, show=False, *params):
    x = copy.deepcopy(look)
    if b2 is None:
        b2 = npy.array([3, 90, 0.8, 0.25, 0.85, 4, 2, 2, 4, 0.25, 2])
    if b1 is None:
        b1 = npy.array([0.5, 30, 0, 0, 0.5, 0.35, 0.2, 0.4, 1.2, 1, 0.4])

    if weights is None:
        weights = npy.array([0.24, 0.12, 0.17, 0.12, 0, 0, 0, 0, 0, 0, 0.35])
    if isinstance(weights,list):
        weights = npy.array(weights)
    if properties is None:
        properties = npy.array(["cent_dist", "angle_diff", "area_ratio", "int_area",
                                "bdelta", "haus", "ph", "med", "msd", "fom", "minsep"])
    if isinstance(properties,list):
        properties = npy.array(properties)
    if show:
        begin_tiid = time.time()

    if (x['grd_ob_features'] is None or 1 not in x['grd_ob_features'].keys()) or \
            (x['grd_fo_features'] is None):
        print("interester: No features in one or both of the fields.  Returning NULL.")
        return None
    zerow = (weights == 0)
    nozerow = ~zerow
    np = npy.sum(nozerow)
    if np == 0:
        print("interester: all weights are zero so that no interest is calculated.  Returning NULL.")
        return None
    if np < weights.size:
        # 过滤零
        # weights = list(filter(lambda we:we!= 0, weights))
        # b1、b2删除weights中索引为零的元素
        properties = properties[nozerow]
        weights = weights[nozerow]
        b1 = b1[nozerow]
        b2 = b2[nozerow]
    a0 = npy.full(np, None)
    a1 = npy.full(np, None)
    # type.ind < - ! is.element(properties, c("area.ratio", "int.area", "fom"))
    type_ind = []
    for i, val in enumerate(properties):
        type_ind.append(val not in ["area_ratio", "int_area", "fom"])
    type_ind = npy.array(type_ind)
    if type_ind.size > 0:
        for i, val in enumerate(type_ind):
            if val:
                a1[i] = -1 / (b2[i] - b1[i])
                a0[i] = 1 - a1[i] * b1[i]
    # type.ind < - is.element(properties, c("area.ratio", "int.area"))
    type_ind = []
    for i, val in enumerate(properties):
        type_ind.append(val in ["area_ratio", "int_area"])
    type_ind = npy.array(type_ind)
    if type_ind.size > 0:
        for i, val in enumerate(type_ind):
            if val:
                a1[i] = 1 / (b2[i] - b1[i])
                a0[i] = 1 - a1[i] * b2[i]

    def ipwlin(x, b1, b2, a0, a1, property, *params):
        #print(x)
        dn = []
        for j, value in enumerate(properties):
            dn.append(value in ["cent_dist", "angle_diff", "bdelta", "haus", "ph", "med", "msd", "minsep"])
        up = []
        for j, value in enumerate(properties):
            up.append(value in ["area_ratio", "int_area"])
        fom = (property == "fom")
        res = npy.zeros(len(x))
        x_is_na = is_na(x)
        if any(x_is_na):
            bool_array = npy.array(dn) * npy.array(x_is_na)
            if any(bool_array):
                for j, value in enumerate(bool_array):
                    if value:
                        x[list(x)[j]] = b2[j]
            bool_array = npy.array(up) * npy.array(x_is_na)
            if any(bool_array):
                for j, value in enumerate(bool_array):
                    if value:
                        x[list(x)[j]] = b1[j]
            if any(fom):
                for j, value in enumerate(fom):
                    if value:
                        x[list(x)[j]] = 100
        if any(dn):
            dn2 = []
            dn3 = []
            for j, value in enumerate(list(x)):
                dn2.append(x[value] <= b1[j])
                dn3.append(x[value] > b2[j])
            bool_array = npy.array(dn) * npy.array(dn2)
            if any(bool_array):
                res[npy.array(bool_array)] = 1
            bool_array = npy.array(dn) * npy.array(dn3)
            if any(bool_array):
                res[npy.array(bool_array)] = 0
            bool_array = npy.array(dn) * ~(npy.array(dn2) + npy.array(dn3))
            if any(bool_array):
                look = []
                for j, value in enumerate(bool_array):
                    if value:
                        look.append(a0[j] + a1[j] * x[list(x)[j]])
                look = npy.array(look)
                if any(look < 0):
                    look[look < 0] = 0
                res[bool_array] = look

        if any(up):
            up2 = []
            up3 = []
            for j, value in enumerate(list(x)):
                up2.append(x[value] < b1[j])
                up3.append(x[value] >= b2[j])
            bool_array = npy.array(up) * npy.array(up2)
            if any(bool_array):
                res[npy.array(bool_array)] = 0
            bool_array = npy.array(up) * npy.array(up3)
            if any(bool_array):
                res[npy.array(bool_array)] = 1
            bool_array = npy.array(up) * ~(npy.array(up2) + npy.array(up3))
            if any(bool_array):
                look = []
                for j, value in enumerate(bool_array):
                    if value:
                        look.append(a0[j] + a1[j] * x[list(x)[j]])
                look = npy.array(look)
                if any(look > 1):
                    look[look > 1] = 1
                res[bool_array] = look
        if any(fom):
            look = b1[fom] * math.exp(-0.5 * ((x[fom] - 1) / b2[fom]) ** 4)
            if look < 0:
                look = 0
            elif look > 1:
                look = 1
            res[fom] = look
        # 赋予name，names(res) < - names(x)
        result = {}
        for j, value in enumerate(list(x)):
            result[value] = res[j]
        #print(result)
        return result

    def ifun(id, bigX, Xhat, b1, b2, a0, a1, p, *params):
        if show:
            print("Forecast feature: " + id[1] + " vs Observed feature: " + id[0] + "\n")
        XtmpAttributes = get_attributes_for_feat(bigX)
        YtmpAttributes = get_attributes_for_feat(Xhat)
        remove_list = ['Type', 'xrange', 'yrange', 'dim', 'xstep', 'ystep', 'warnings', 'xcol', 'yrow', 'area']
        xkeys = remove_key_from_list(list(bigX.keys()), remove_list)
        ykeys = remove_key_from_list(list(Xhat.keys()), remove_list)
        # Xtmp = {"m": bigX[xkeys[id[0]]]}
        # Ytmp = {"m": Xhat[ykeys[id[1]]]}
        Xtmp = {"m": bigX[id[0]]}
        Ytmp = {"m": Xhat[id[1]]}
        Xtmp.update(XtmpAttributes)
        Ytmp.update(YtmpAttributes)
        #A = feature_comps(grd_fo=Ytmp, grd_ob=Xtmp, which_comps=p)
        A = feature_comps(look,label_ob=id[0],label_fo=id[1], which_comps=p)
        nomen = A.keys()
        res = ipwlin(A, b1=b1, b2=b2, a0=a0, a1=a1, property=p)
        if "angle_diff" in p:
            #fax = feature_axis(Xtmp)
            #fay = feature_axis(Ytmp)
            fax = feature_axis(look,label=id[0],ob_or_fo="ob")
            fay = feature_axis(look,label=id[1],ob_or_fo="fo")
            if fax is None or fay is None:
                con = 0
            else:
                aspX = fax['aspect_ratio']
                aspY = fay['aspect_ratio']
                conX = ((aspX - 1) ** 2 / (aspX ** 2 + 1)) ** 0.3
                conY = ((aspY - 1) ** 2 / (aspY ** 2 + 1)) ** 0.3
                con = math.sqrt(conX * conY)
            res["angle_diff"] = res["angle_diff"] * con
        # if "cent_dist" in p and "area_ratio" in p:
        #     # 子集 all( is.element(c("cent.dist", "area.ratio"), p))
        #     res["cent_dist"] = res["cent_dist"] * res["area_ratio"]
        return res

    #N = len(x['grd_ob_features'].keys()) - 11
    #M = len(x['grd_fo_features'].keys()) - 11
    #ind = npy.stack((npy.tile(npy.arange(N), M), (npy.arange(M)).repeat(N)), axis=-1)
    label_list_ob = npy.array(look["label_list_ob"])
    label_list_fo = npy.array(look["label_list_fo"])
    N = len(label_list_ob)
    M = len(label_list_fo)
    if N ==0 or M ==0:
        return None
    ind = npy.stack((npy.tile(label_list_ob, M),label_list_fo.repeat(N)), axis=-1)
    if show:
        print(
            "\n\nFinding interest between each pair of " + N + " observed features and " + M + " forecast features.\n\n")

    res1 = npy.apply_along_axis(ifun, axis=1, arr=ind, bigX=x['grd_ob_features'], Xhat=x['grd_fo_features'],
                                b1=b1, b2=b2, a0=a0, a1=a1, p=properties, *params)
    res_array_length_x = len(res1[0])
    res_array = npy.empty([0, res_array_length_x])
    for i in range(len(res1)):
        res_inside = npy.zeros((1, res_array_length_x))
        for j, val in enumerate(list(res1[i])):
            res_inside[0, j] = res1[i][val]
        res_array = npy.append(res_array, res_inside, axis=0)

    res_array_interest = pd.DataFrame(res_array, columns = properties.tolist())
    interest = (npy.nan_to_num(res_array_interest)).reshape(M,N,len(properties))
    coords = {"fo_label_id":npy.arange(M)+1,"ob_label_id":npy.arange(N)+1,"properties":properties.tolist()}

    #print(type(coords))
    #out = {'interest': res_array_interest,
           #'total_interest': npy.sum(res_array * npy.tile(weights, (N * M, 1)), axis=1).reshape((N, M))}
    out = {"properties":properties.tolist(),
          'interest': interest,
           'total_interest': npy.array(npy.sum(npy.nan_to_num(res_array * npy.tile(weights, (N * M, 1))), axis=1).reshape((M, N))),
           "label_list_ob": label_list_ob,
           "label_list_fo": label_list_fo,
           }
    # a = attributes(x)
    # a['names'] = None
    # attributes(out) = a
    # out['total_interest'] = matrix(colSums(res1 * matrix(weights, np, N * M), na.rm = TRUE), N, M)

    if show:
        print(time.time() - begin_tiid)
    return out


def is_na(x):
    res = []
    for i, val in enumerate(list(x)):
        res.append(x[val] is None)
    return res

'''
if __name__ == '__main__':
    data = npy.load("../../deltammResult_PA2.npy", allow_pickle=True).tolist()
    z = interester(data)
    print("hello")
'''


from . import as_unitname
from . import datagrabber_spatialVx
from . import deltametric
from . import distmap
from . import im
from . import loc_list_setup
from . import locperf
from . import make_spatialVx
from . import solutionset





import  numpy as np
from sklearn.utils import resample
from scipy import stats
import pandas as pd
import meteva
import copy


def CI_fun(CI_field, est_field, replacement=np.nan):
    test = est_field[0.5 * CI_field < abs(est_field)] = replacement
    return test


def inside(DF):
    # result = CI_fun(DF['Upper'] - DF['Lower'], DF['Estimate'])
    aa = DF['Upper'] - DF['Lower']
    bb = DF['Estimate']
    result = 0.5 * aa < abs(bb)  #满足条件代表平均值是显著偏离1的
    return result


def sig_coverage(DF):
    out = {}
    tmp = inside(DF)
    #sum(DF['Estimate'] == None) =0 表示df中的内容是误差均值，否则表示其它暂无支持的统计内容
    # out 是表示均值显著的占比。
    out = sum(tmp[tmp == 1]) / len(tmp) - sum(DF['Estimate'] == None)
    return out


def sig_cor_t(r, Len=40):
    t = abs(r) * np.sqrt((Len - 2) / (1 - r ** 2))
    # qt -> stats.t(df=359).ppf((0.84473448, 0.15382939))
    palpha_cor = 1 - stats.t.cdf(t, df=Len - 2)  # t分布的分布函数
    return palpha_cor


def MCdof(x, ntrials=1000, field_sig=0.05, zfun="rnorm", zfun_args='NULL',
          which_test=["t", "Z", "cor.test"], show=False):

    output = {}
    if len(which_test) > 1:
        which_test = "t"
    x = np.array(x)
    xdim = x.shape
    tlen = xdim[0]
    B_dof_test = np.zeros(ntrials)
    if show:
        print("\n", "Looping through ", ntrials, " times to simulate data and take correlations.  Enjoy!\n")
    for i in range(1, ntrials):
        if show and (i < 100 or i % 100 == 0):
            print(i, " ")
        # rnorm(n, mean = 0, sd = 1)返回值是n个正态分布随机数构成的向量。
        mu, sigma = 0, 1  # mean and standard deviation
        z = np.random.normal(mu, sigma, tlen)  # 正态分布存在差异
        '''
        #绘制正态分布图
        count, bins, ignored = plt.hist(s, 30, density=True)
        plt.plot(bins, 1/(sigma * np.sqrt(2 * np.pi)) * np.exp( - (bins - mu)**2 / (2 * sigma**2) ),
                 linewidth=2, color='r')
        plt.show()
        '''
        # 判断为否，不执行
        tmp = np.array([])
        if which_test == "cor.test":
            # tmp = apply(x, 2, cortester, y = z)
            tmp = np.corrcoef(x)[1, 0]
        else:
            #默认进行t检验
            # 计算pearson相关系数
            cor = np.array([])
            for j in range(x.shape[1]):  # 循环变量有问题
                x_j = x[:, j]
                xz = np.array([x_j, z])
                cor_0 = np.corrcoef(xz)[1, 0]  # 目前存在偏差是由于z的结果不一致导致，方法本身经过验证是没问题的
                cor = np.append(cor, cor_0)
            cor_value = np.abs(cor)
            # 只翻译默认的第一个参数
            if which_test == "t":
                tmp = sig_cor_t(r=cor_value, Len=tlen)  # 输出结果
            '''
            else if (which.test == "Z") 
                tmp <- sig.cor.Z(cor.value, len = tlen, ...)
            '''

        tmp_small = tmp[field_sig>tmp] #提取显著性超阈值那部分显著性

        # if tmp_small.size == 0:
        #     B_dof_test[i] = 1
        # else:
        #     B_dof_test[i] = np.mean(tmp_small)  # 显著性超阈值部分的均值
        B_dof_test[i] = tmp_small.size / tmp.size

    minsigcov_val = np.quantile(a=B_dof_test, q=1 - field_sig)
    output = {'MCprops': B_dof_test, 'minsigcov': minsigcov_val}
    return output

def is_sig(X, blockboot_result_df, n=1000, fld_sig=0.05, show=False):
    output = {}
    '''
    X = errfield
    blockboot_result_df = hold
    n = ntrials
    fld_sig = field_sig
    '''

    # sig_result 表示利用蒙特卡罗方法测试所有格点和随机数求相关后，表现为具有显著性的格点数占比。
    sig_result = MCdof(X, ntrials=n, field_sig=0.05, show=show)['minsigcov']

    #actual_coverage表示平均值具有显著性的格点数占比。
    actual_coverage = sig_coverage(DF=blockboot_result_df)
    sig = actual_coverage > sig_result  #输出结果
    '''
    output <- list(name = as.character(deparse(substitute(X))), 
                   required = as.numeric(sig.results), actual = as.numeric(actual.coverage), 
                   issig = as.logical(sig))
    '''
    output = {"name": X, "required": sig_result, "actual": actual_coverage, "issig": sig}

    return output


def mean_estimate(Z, numrep=1000, alpha=0.05,block_length='NULL', bootfun="mean", show=False):

    booted = {}
    if bootfun == 'mean':
        Z = np.ma.masked_array(Z, np.isnan(Z))
        bootfun = np.mean(Z, axis=0)
        booted_t0 = bootfun
    Z = np.array(Z)
    if len(Z.shape) == 1:
        Z = Z.reshape((Z.size,1))
    zdim = Z.shape

    n = zdim[0]
    m = zdim[1]
    out = pd.DataFrame(np.zeros(shape=(m, 3)), columns=['Lower', 'Estimate', 'Upper'])
    if block_length is 'NULL':
        block_length = np.floor(np.sqrt(n))

    if block_length == 1:
        # 判断为假，不执行
        # booted = boot(Z, bootfun, R = numrep)    #未翻译
        for i in range(Z.shape[1]):
            booted_list = [np.mean(resample(Z[:, i], n_samples=Z.shape[0])) for _ in
                           range(numrep)]  # n_samples是受tseries控制
            booted_0 = np.array(booted_list)
            booted = np.append(booted, booted_0)
    else:
        # booted = tsboot(Z, bootfun, l = block.length, R = numrep, sim = "fixed")
        booted = np.array(())
        for i in range(Z.shape[1]):
            booted_list = [np.mean(resample(Z[:, i], n_samples=Z.shape[0])) for _ in range(numrep)]  # n_samples是受tseries控制
            booted_0 = np.array(booted_list)
            booted = np.append(booted, booted_0)
            # booted = np.stack((booted, booted_0), axis = 1)
    booted_t = booted.reshape(numrep, Z.shape[1], order='F')
    booted = {'t0': booted_t0, 't': booted_t, 'R': numrep}
    out['Estimate'] = booted['t0']
    #print(booted_t)
    if block_length == 1:
        # 判断为假,不执行
        conf = 1 - alpha
        tmp_Lower = np.percentile(a=booted_t, q=100*(1 + conf) / 2, axis=0)
        tmp_Upper = np.percentile(a=booted_t, q=100*(1 - conf) / 2, axis=0)
        out['Lower'] = tmp_Lower
        out['Upper'] = tmp_Upper

        '''
        #原本R的逻辑
        for i in range(1, m):
            tmp = boot_CI(booted, conf = 1-alpha, Type = "bac", index = i)
            out['Lower'][i] = tmp["bca"][:, 4]
            out['Upper'][i] = tmp["bca"][:, 5]
        '''
    else:
        if show:
            print(
                "LocSig: You chose to use the BCa method, but block.length != 1.  Using percentile method with circular block bootstrap instead.")

        conf = 1 - alpha
        tmp_Lower = np.percentile(a=booted_t, q=100*(1 - conf) / 2, axis=0)
        tmp_Upper = np.percentile(a=booted_t, q=100*(1 + conf) / 2, axis=0)
        out['Lower'] = tmp_Lower
        out['Upper'] = tmp_Upper

        '''
        #原本R的逻辑
        for i in range(1, m):
            #计算置信度区间（分位数:np.percentile）
            #可信度=0.90,置信区间宽度=0.05, 置信区间上限=0.07,置信区间下限=0.02,表示:发生在区间(0.02,0.07)这个范围内的可能性为0.90
            tmp = boot_CI(booted, conf = 1-alpha, Type = "perc", index = i)
            out['Lower'[i]] = tmp["prec"][:, 4]
            out['Upper'][i] = tmp["prec"][:, 5]
        '''

    # out['class'] = "LocSig"
    #result = np.array([out["Lower"],out["Estimate"],out["Upper"]]).T

    return out


def local_sig(ob,fo,numrep=1000, alpha=0.05,block_length='NULL', bootfun="mean", show=False):


    if len(fo.shape) == len(ob.shape):
         error = fo - ob
         out = mean_estimate(error,numrep=numrep, alpha=alpha,block_length=block_length, bootfun=bootfun, show=show)
         index = np.abs(out["Estimate"].values) - (out["Upper"].values-out["Lower"].values)/2

    else:
        index_list = []
        for i in range(fo.shape[0]):
            error = fo[i,...] - ob
            out = mean_estimate(error, numrep=numrep, alpha=alpha, block_length=block_length, bootfun=bootfun,
                                show=show)
            index = np.abs(out["Estimate"].values) - (out["Upper"].values - out["Lower"].values) / 2
            index_list.append(index)
        index = np.array(index_list)
    return index

def spatbiasFS(X, Y, loc='NULL', block_length='NULL', alpha_boot=0.05,
               field_sig=0.05, bootR=1000, ntrials=1000, show=False):
    out = {}
    '''
    if (!is.null(loc)) {
        data.name <- c(as.character(substitute(X)), as.character(substitute(Y)), 
                   as.character(substitute(loc)))
        names(data.name) <- c("verification", "forecast", "locations")
        }
    else {
        data.name <- c(as.character(substitute(X)), as.character(substitute(Y)))
        names(data.name) <- c("verification", "forecast")
        }
    out$data.name <- data.name
    '''
    errfield = Y - X
    hold = mean_estimate(Z=errfield, numrep=bootR, block_length=block_length,
                         alpha=alpha_boot)
    res = is_sig(errfield, hold, n=ntrials, fld_sig=field_sig, show=show)
    out['block_boot_results'] = hold
    out['sig_results'] = res
    out['field_significance'] = field_sig
    out['alpha_boot'] = alpha_boot
    out['bootR'] = bootR
    out['ntrials'] = ntrials
    out['class'] = "spatbiasFS"

    return out


def field_sig(ob, fo, member_list = None,numrep=1000,alpha=0.05, block_length='NULL',return_mid = False):

    index_list = []
    if len(fo.shape) == len(ob.shape):
        index = local_sig(ob,fo, numrep=numrep, alpha=alpha,block_length=block_length)
        rate0 = len(index[index>0])/len(index)
        errfield = fo - ob
        rate1 = MCdof(errfield, ntrials=numrep, field_sig=0.05)['minsigcov']
        index_dict = {"sig_rate":rate0,"rate_threshold":rate1,"is_sig":rate0>=rate1}
        index_list.append(index)
    else:
        index_dict = {"sig_rate": [], "rate_threshold": [], "is_sig": []}
        for i in range(fo.shape[0]):
            fo1= fo[i, ...]
            index = local_sig(ob, fo1, numrep=numrep, alpha=alpha, block_length=block_length)
            rate0 = len(index[index > 0]) / len(index)
            errfield = fo1- ob
            rate1 = MCdof(errfield, ntrials=numrep, field_sig=0.05)['minsigcov']
            index_dict["sig_rate"].append(rate0)
            index_dict["rate_threshold"].append(rate1)
            index_dict["is_sig"].append(rate0>=rate1)
            index_list.append(index)

    index_f =pd.DataFrame(index_dict)
    if member_list is not None:
        index_f.index = member_list

    if return_mid:
        index_l = np.array(index_list)
        return index_f,index_l
    else:
        return index_f

def field_sig_id(sta_all,s = None,numrep=1000,alpha=0.05,plot = None,save_path = None,show = False, block_length='NULL',**kwargs):

    sta_all1 = meteva.base.sele_by_dict(sta_all, s=s)

    level = set(sta_all1["level"].values.tolist())
    if len(level)>1:
        print("暂时不支持同时多个层次开展显著性检验，可以先用s参数选取某一层的数据进行检验")

    sta_with_IV = sta_all1.drop_duplicates("id")
    index_station = meteva.base.in_member_list(sta_with_IV,member_list=[1,2],name_or_index="index")
    nid = len(sta_with_IV.index)
    sta_with_IV["time"] =meteva.base.IV
    sta_with_IV["dtime"] = meteva.base.IV
    sta_with_IV = meteva.base.in_member_list(sta_with_IV,[0],name_or_index="index")

    for i in range(1,3):
        sta_expand = []
        value_list = list(set(sta_all1.iloc[:,i].values.tolist()))
        if i == 1:
            for j in range(len(value_list)):
                value_list[j] = meteva.base.tool.all_type_time_to_time64(value_list[j])
        for value in value_list:
            sta1 = copy.deepcopy(sta_with_IV)
            sta1.iloc[:,i] = value
            sta_expand.append(sta1)
        sta_with_IV = meteva.base.concat(sta_expand)

    sta_combine = meteva.base.combine_on_level_time_dtime_id(sta_all1, sta_with_IV,how = 'outer')
    sta_combine = sta_combine.fillna(value=0)
    sta_combine.sort_values(by=["time","dtime","id"])
    nt = int(len(sta_combine.index)/nid)
    data_names  = meteva.base.get_stadata_names(sta_combine)
    nmember = len(sta_combine.columns)
    #sta_ob = meteva.base.sele_by_para(sta_combine,member=[data_names[0]])
    obs = sta_combine.iloc[:,6].values.reshape((nt,nid))
    fst = sta_combine.iloc[:,7:nmember-1].values.reshape(nmember-8,nt,nid)

    index_f,index_l = field_sig(obs,fst,member_list=data_names[1:-1],numrep = numrep,alpha = alpha,block_length=block_length,return_mid=True)


    for i in range(len(data_names)-2):
        index_station.iloc[:,6+i] = index_l[i,:]
    if plot == "scatter":
        if save_path is None:
            show = True
        meteva.base.tool.plot_tools.scatter_sta(index_station,cmap="me",subplot="member",sup_title="单点偏差显著性指标",show=show,save_path=save_path,**kwargs)

    return index_f,index_station




import math
import meteva
from meteva.method.space.fuzzy_logic.lib.kernel2dsmooth import kernel2dsmooth


def  hoods2dsmooth_grd(grd,half_window_size):
    '''
    以网格形式数据作为输入的平滑函数
    :param grd:
    :param level:
    :return:
    '''
    level = half_window_size *2+1
    grid0 = meteva.base.get_grid_of_data(grd)
    x = grd.values.squeeze()
    level_w = hoods2dsmooth(x, level, None, True)
    s_px = hoods2dsmooth(x, level, level_w)
    sm_x = meteva.base.grid_data(grid0, s_px)
    return sm_x



def hoods2dsmooth(x, lambd, w=None, setup=False, **args):
    if math.floor(lambd) != lambd:
        print("hoods2dsmooth: attempting to give an illegal value for the neighborhood length.  Flooring lambd.")
        lambd = math.floor(lambd)

    if lambd % 2 == 0:
        print("hoods2dsmooth: attempting to give an even neighborhood length, subtracting one from it.")
        lambd = lambd - 1

    if lambd < 1:
        print("hoods2dsmooth: attempting to give an illegal value for the neighborhood length.  Setting to one, "
              "and returning x untouched.")
        lambd = 1

    if lambd == 1:
        return x
    else:
        return kernel2dsmooth(x=x, kernel_type="boxcar", n=lambd, w=w, setup=setup, **args)



#!/usr/bin/python
# -*- coding: utf-8 -*-
#Filename: ftpconn.py

'''
Created on Oct 13, 2020

@author: anduin
'''

from ftplib import FTP
from urllib.parse import urlparse
from urllib.parse import urljoin
from urllib.parse import urlencode
import socket

try:
    import ssl
except ImportError:
    _SSLSocket = None
else:
    _SSLSocket = ssl.SSLSocket

from .loglib import *

class FtpConn():
    @staticmethod
    def joinUrl(url, path):
        return urljoin(url, path)
    
    @staticmethod
    def getPath(url):
        try:
            if url.lower().find(r'ftp://') == 0:
                return urlparse(url).path
            else:
                return url
        except Exception as data:
            LogLib.logger.error('getPath except ' + str(data))
            #print(data)
            return None

    def __init__(self, host, port=21):
        """ 初始化 FTP 客户端
        参数:
            host:ip地址
            port:端口号
        """
        # print("__init__()---> host = %s ,port = %s" % (host, port))
        self.host = host
        self.port = port
        self.ftp = FTP()
        # 重新设置下编码方式
        self.ftp.encoding = 'gbk'
        self.file_list = []
        self.default_bufsize = 1024*64

    def login(self, username, password, pasv=False):
        """ 初始化 FTP 客户端
            参数:
                  username: 用户名

                 password: 密码
            """
        try:
            timeout = 60
            socket.setdefaulttimeout(timeout)
            # 0主动模式 1 #被动模式
            LogLib.logger.debug('login set_pasv %s ' % str(pasv))
            self.ftp.set_pasv(pasv)
            # 打开调试级别2，显示详细信息
            # self.ftp.set_debuglevel(2)

            #print('开始尝试连接到 %s' % self.host)
            LogLib.logger.debug('login connect %s  %d' % (self.host, self.port))
            self.ftp.connect(self.host, self.port)
            #print('成功连接到 %s' % self.host)

            #print('开始尝试登录到 %s' % self.host)
            LogLib.logger.debug('login login %s' % (username))
            self.ftp.login(username, password)
            #print('成功登录到 %s' % self.host)

            #print(self.ftp.welcome)
            LogLib.logger.debug('login over %s %s' % (self.host, username))

            return True
        except Exception as data:
            #self.deal_error("FTP 连接或登录失败 ，错误描述为：%s" % data)
            LogLib.logger.error('login except %s %s %s' % (self.host, username, str(data)))
            
            return False
        
    def quit(self):
        try:
            self.ftp.quit()
        except:
            pass

    def get_rfile_size(self, remote_file):
        """获取远程文件大小
           参数:
             remote_file: 远程文件
             如果remote_file是全路径ftp://abc.com/ddd/dd；
             如果remote_file是相对路径，必须是相对于当前路径abc/aa，或者根路径/abc/aa。
             如果remote_file仅仅是文件名，当前路径必须是该文件所在的目录；
        """
        try:
            fpath = FtpConn.getPath(remote_file)
            if fpath is None:
                return -2

            self.ftp.voidcmd('TYPE I')
            return self.ftp.size(fpath)
        except Exception as sdata:
            #print("get_rfile_size() 错误描述为：%s" % sdata)
            LogLib.logger.error('get_rfile_size except %s %s' % (remote_file, str(sdata)))

            return -1

    def is_same_size(self, local_file, remote_file):
        """判断远程文件和本地文件大小是否一致

           参数:
             local_file: 本地文件
             remote_file: 远程文件
        """
        remote_file_size = 0
        local_file_size = 0

        try:
            remote_file_size = self.get_rfile_size(remote_file) if type(remote_file) is str else remote_file
        except Exception as sdata:
            # print("is_same_size() 错误描述为：%s" % sdata)
            LogLib.logger.error('is_same_size except %s %s' % (remote_file, str(sdata)))

            return -1
            #remote_file_size = -1

        try:
            local_file_size = os.path.getsize(local_file) if os.path.exists(local_file) else -1
        except Exception as fdata:
            # print("is_same_size() 错误描述为：%s" % fdata)
            LogLib.logger.error('is_same_size except %s %s' % (local_file, str(sdata)))

            return -2
            #local_file_size = -1

        #print('local_file_size:%d  , remote_file_size:%d' % (local_file_size, remote_file_size))
        if remote_file_size == local_file_size:
            return 1
        else:
            return 0

    def download_file(self, local_file, remote_file):
        """从ftp下载文件
            参数:
                local_file: 本地文件
                remote_file: 远程文件
        """
        #print("download_file()---> local_path = %s ,remote_path = %s" % (local_file, remote_file))
        LogLib.logger.info('download_file start %s %s' % (local_file, remote_file))
        nrst = self.is_same_size(local_file, remote_file)
        if nrst > 0:
            #print('%s 文件大小相同，无需下载' % local_file)
            LogLib.logger.info('download_file 文件大小相同，无需下载 %s %s' % (local_file, remote_file))
            return True
        elif nrst < 0:
            return False
        else:
            try:
                #print('>>>>>>>>>>>>下载文件 %s ... ...' % local_file)
                LogLib.logger.info('download_file downloading %s %s' % (local_file, remote_file))
                buf_size = self.default_bufsize
                with open(local_file, 'wb') as file_handler:
                    self.ftp.retrbinary('RETR %s' % remote_file, file_handler.write, buf_size)
                #file_handler.close()

                LogLib.logger.info('download_file download over %s %s' % (local_file, remote_file))

                return True
            except Exception as data:
                #print('下载文件出错，出现异常：%s ' % err)
                LogLib.logger.error('download_file except %s %s %s' % (local_file, remote_file, str(data)))

                return False
            
    def download_file_to_buf(self, remote_file):
        """从ftp下载文件内容，不保存文件，将数据直接返回，不建议处理大文件
            参数:
                remote_file: 远程文件
        """
        #print("download_file_to_buf()---> remote_path = %s" % (remote_file))
        LogLib.logger.info('download_file_to_buf start %s' % (remote_file))
        fsize = self.get_rfile_size(remote_file)
        if fsize < 0:
            LogLib.logger.error('download_file_to_buf get_rfile_size %s' % (remote_file))
            return False, None
        elif fsize == 0:
            return True, None
        else:
            try:
                #print('>>>>>>>>>>>>下载文件 %s ... ...' % local_file)
                LogLib.logger.info('download_file_to_buf downloading %s' % (remote_file))
                self.ftp.voidcmd('TYPE I')

                down_size = 0
                down_data = None
                with self.ftp.transfercmd('RETR %s' % remote_file, 0) as conn:
                    while down_size < fsize:
                        buf_size = self.default_bufsize
                        if fsize - down_size < buf_size:
                            buf_size = fsize - down_size

                        rdata = conn.recv(buf_size)
                        if rdata:
                            down_size += len(rdata)
                            if down_data is None:
                                down_data = rdata
                            else:
                                down_data += rdata
                        else:
                            break
                        
                    # shutdown ssl layer
                    if _SSLSocket is not None and isinstance(conn, _SSLSocket):
                        conn.unwrap()

                resp = self.ftp.voidresp()

                if fsize != down_size:
                    LogLib.logger.error('download_file_to_buf download over,but size error %s %s' % (remote_file, str(resp)))
                    return False, None
                else:
                    LogLib.logger.info('download_file_to_buf download over %s %s' % (remote_file, str(resp)))
                    return True, down_data
            except Exception as data:
                #print('下载文件出错，出现异常：%s ' % err)
                LogLib.logger.error('download_file_to_buf except %s %s' % (remote_file, str(data)))

                return False, None

    def download_file_tree(self, local_path, remote_path):
        """从远程目录下载多个文件到本地目录
                       参数:
                         local_path: 本地路径

                         remote_path: 远程路径
                """
        print("download_file_tree()--->  local_path = %s ,remote_path = %s" % (local_path, remote_path))
        try:
            self.ftp.cwd(remote_path)
        except Exception as err:
            print('远程目录%s不存在，继续...' % remote_path + " ,具体错误描述为：%s" % err)
            return

        if not os.path.isdir(local_path):
            print('本地目录%s不存在，先创建本地目录' % local_path)
            os.makedirs(local_path)

        print('切换至目录: %s' % self.ftp.pwd())

        self.file_list = []
        # 方法回调
        self.ftp.dir(self.get_file_list)

        remote_names = self.file_list
        print('远程目录 列表: %s' % remote_names)
        for item in remote_names:
            file_type = item[0]
            file_name = item[1]
            local = os.path.join(local_path, file_name)
            if file_type == 'd':
                print("download_file_tree()---> 下载目录： %s" % file_name)
                self.download_file_tree(local, file_name)
            elif file_type == '-':
                print("download_file()---> 下载文件： %s" % file_name)
                self.download_file(local, file_name)
            self.ftp.cwd("..")
            print('返回上层目录 %s' % self.ftp.pwd())
        return True

    def upload_file(self, local_file, remote_file_catalog,record_abspath,remote_file_catalog_r):
        """从本地上传文件到ftp
           参数:
             local_path: 本地文件
             remote_path: 远程文件
        """
        exe_status = {}
        # 创建一个字典表，key是文件时次,values是文件名称。
        if os.path.exists(record_abspath):
            with open(record_abspath, 'r') as f:
                for line in f.readlines():
                    line = line.strip()
                    #print("line:{}".format(line))
                    shici,filenames_local,wenjian_time,upload_time = line.split(',')
                    exe_status[shici] = filenames_local
        #上传的文件名称
        filenames = local_file.split('/')[-1]
        times = filenames.split(".")[1]
        print(filenames,times)
        #if not os.path.exists(remote_file):
        #    self.ftp.mkd(remote_file)
        remote_file = os.path.join(remote_file_catalog,filenames)
        print(remote_file,local_file)
        if not os.path.isfile(local_file):
            print('%s 不存在' % local_file)
            return
        #需要修改代码，做成是否能够对应上字典表
        # 日志记录中有对应文件执行信息不再执行
        if filenames in exe_status.get(times, ''):
            print("已经存在")
            return
        
        #print("目录文件：{}".format(self.ftp.nlst(remote_file_catalog_r)))
        #print(remote_file_catalog)
        try:
            if remote_file_catalog in self.ftp.nlst(remote_file_catalog_r):
                pass
            else:
                self.ftp.mkd(remote_file_catalog)
        except:
            self.ftp.mkd(remote_file_catalog_r)
        #if not os.path.exists(local_file):
        #    os.makedirs(local_file)
        buf_size = 1024
        #获取文件时间
        t = os.path.getctime(local_file)
        file_time = str(datetime.datetime.fromtimestamp(t)).split('.')[0]
        #print(dir_t,type(dir_t))
        #file_time = time.strptime(dir_t, "%Y-%m-%d %H:%M:%S")
        file_handler = open(local_file, 'rb')
        self.ftp.storbinary('STOR %s' % remote_file, file_handler, buf_size)
        print("上传成功")
        file_handler.close()
        #上传文件时间
        time_now = time.localtime()
        date_now = str(time.strftime('%Y-%m-%d %H:%M:%S', time_now)).split(".")[0]
        text = times + ',' + filenames + ',' + file_time + ',' + date_now + '\n'
        #将日志信息写进log里边。
        print(record_abspath)
        with open(record_abspath, 'a') as f_w:
            f_w.write(text)
        #print(text)

    def upload_file_tree(self, local_path, remote_path):
        """从本地上传目录下多个文件到ftp
           参数:

             local_path: 本地路径

             remote_path: 远程路径
        """
        if not os.path.isdir(local_path):
            #print('本地目录 %s 不存在' % local_path)
            return
        """
        创建服务器目录
        """
        try:
            self.ftp.cwd(remote_path)  # 切换工作路径
        except Exception as e:
            base_dir, part_path = self.ftp.pwd(), remote_path.split('/')
            part_path.append("00")
            for p in part_path[1:-1]:
                base_dir = base_dir + p + '/'  # 拼接子目录
                #print("拼接子目录")
                print(base_dir)
                try:
                    self.ftp.cwd(base_dir)  # 切换到子目录, 不存在则异常
                except Exception as e:
                    print('INFO:', e)
                    self.ftp.mkd(base_dir)  # 不存在创建当前子目录
                    zi_dir = base_dir + '00'+ '/'
                    self.ftp.mkd(zi_dir)
        #self.ftp.cwd(remote_path)
        print('切换至远程目录: %s' % self.ftp.pwd())

        local_name_list = os.listdir(local_path)
        print('本地目录list: %s' % local_name_list)
        #print('判断是否有服务器目录: %s' % os.path.isdir())

        for local_name in local_name_list:
            src = os.path.join(local_path, local_name)
            print("src路径=========="+src)
            if os.path.isdir(src):
                try:
                    self.ftp.mkd(local_name)
                except Exception as err:
                    print("目录已存在 %s ,具体错误描述为：%s" % (local_name, err))
                print("upload_file_tree()---> 上传目录： %s" % local_name)
                print("upload_file_tree()---> 上传src目录： %s" % src)
                self.upload_file_tree(src, local_name)
            else:
                print("upload_file_tree()---> 上传文件： %s" % local_name)
                self.upload_file(src, local_name)
        self.ftp.cwd("..")

    def close(self):
        """ 退出ftp
        """
        #print("close()---> FTP退出")
        self.ftp.quit()

    def get_file_list(self, line):
        """ 获取文件列表
            参数：
                line：
        """
        file_arr = self.get_file_name(line)
        # 去除  . 和  ..
        if file_arr[1] not in ['.', '..']:
            self.file_list.append(file_arr)

    def get_file_name(self, line):
        """ 获取文件名
            参数：
                line：
        """
        pos = line.rfind(':')
        while (line[pos] != ' '):
            pos += 1
        while (line[pos] == ' '):
            pos += 1
        file_arr = [line[0], line[pos:]]
        return file_arr
if __name__ == '__main__':
    import os
    workdir = os.path.dirname(__file__)
    workdir = os.path.dirname(workdir)
    logdir = os.path.join(workdir, 'testftpconn')
    if not os.path.exists(logdir):
        os.mkdir(logdir)

    LogLib.init()
    LogLib.addTimedRotatingFileHandler(os.path.join(logdir, 'testftpconn.log'))
    LogLib.logger.setLevel(logging.DEBUG)
    
    ftpobj = FtpConn('ftp.sjtu.edu.cn')
    if ftpobj.login('', ''):
        rst, fdata = ftpobj.download_file_to_buf(r'/pub/software/vmlinuz')
        if rst and fdata is not None:
            with open(r'c:\workdoc\a.test', 'wb') as f:
                f.write(fdata)

    LogLib.uninit()

    print('test done')